import os
import pathlib
from nipype.pipeline import engine as pe
from nipype.interfaces import utility as niu
from nipype.interfaces.io import DataSink
from nipype.interfaces.utility import Function
from .inho_correction import init_inho_correction_wf
from .commonspace_reg import init_commonspace_reg_wf,inherit_unbiased_files
from .bold_main_wf import init_bold_main_wf
from .utils import prep_bids_iter, resample_template
from . import preprocess_visual_QC
from .prep_input import init_prep_input_wf,match_iterables

def init_main_wf(data_dir_path, output_folder, opts, name='main_wf'):
    '''
    This workflow organizes the entire processing.

    **Parameters**

        data_dir_path
            Path to the input data directory with proper BIDS folder structure.
        output_folder
            path to output folder for the workflow and datasink
        opts
            parser options for preprocess
        cr_opts
            parser options for confound_correction
        analysis_opts
            parser options for analysis

    **Outputs**


        input_bold
            Input EPIs to the preprocessing
        commonspace_resampled_template
            the anatomical commonspace template after initial resampling
        anat_preproc
            Preprocessed anatomical image after bias field correction and denoising
        anat_mask
            Brain mask inherited from the common space registration
        WM_mask
            Eroded WM mask inherited from the common space registration
        CSF_mask
            Eroded CSF mask inherited from the common space registration
        initial_bold_ref
            Initial EPI median volume subsequently used as 3D reference EPI volume
        inho_cor_bold
            3D reference EPI volume after bias field correction
        bold_to_anat_affine
            affine transform from the EPI space to the anatomical space
        bold_to_anat_warp
            non-linear transform from the EPI space to the anatomical space
        bold_to_anat_inverse_warp
            inverse non-linear transform from the EPI space to the anatomical space
        inho_cor_bold_warped2anat
            Bias field corrected 3D EPI volume warped to the anatomical space
        native_corrected_bold
            Preprocessed EPI resampled to nativespace
        corrected_bold_ref
            3D ref EPI volume from the native EPI timeseries
        motion_params_csv
            .csv file with measured motion timecourses, used as regressors for confound
            correction: 6 rigid body motion parameters + their first temporal derivate 
            + the 12 parameters squared (24 motion parameters)
        FD_voxelwise
            Voxelwise framewise displacement (FD) measures that can be integrated
            to future confound regression.
            These measures are computed from antsMotionCorrStats.
        pos_voxelwise
            Voxel distancing across time based on rigid body movement parameters,
            which can be integrated for a voxelwise motion regression
            These measures are computed from antsMotionCorrStats.
        FD_csv
            .csv file with global framewise displacement (FD) measures
        bold_brain_mask
            EPI brain mask for native corrected bold
        bold_WM_mask
            EPI WM mask for native corrected bold
        bold_CSF_mask
            EPI CSF mask for native corrected bold
        commonspace_bold
            Motion and SDC-corrected EPI timeseries resampled into common space
            by applying transforms from the anatomical common space registration
        commonspace_mask
            EPI brain mask for commonspace bold
        commonspace_WM_mask
            EPI WM mask for commonspace bold
        commonspace_CSF_mask
            EPI CSF mask for commonspace bold
        commonspace_vascular_mask
            EPI vascular mask for commonspace bold
        std_filename
            temporal STD map of the preprocessed timeseries
        tSNR_filename
            temporal SNR map of the preprocessed timeseries
    '''

    workflow = pe.Workflow(name=name)

    # set output node
    outputnode = pe.Node(niu.IdentityInterface(
        fields=['input_bold', 'commonspace_resampled_template', 'anat_preproc', 'initial_bold_ref', 'inho_cor_bold', 'bold_to_anat_affine',
                'bold_to_anat_warp', 'bold_to_anat_inverse_warp', 'inho_cor_bold_warped2anat', 'native_bold', 'native_bold_ref', 'motion_params_csv',
                'FD_voxelwise', 'pos_voxelwise', 'FD_csv', 'native_brain_mask', 'native_WM_mask', 'native_CSF_mask', 'native_vascular_mask',
                'commonspace_bold', 'commonspace_mask', 'commonspace_WM_mask', 'commonspace_CSF_mask', 'commonspace_vascular_mask',
                'std_filename', 'tSNR_filename', 'boldspace_brain_mask']),
        name='outputnode')

    # Datasink - creates output folder for important outputs
    bold_datasink = pe.Node(DataSink(base_directory=output_folder,
                                     container="bold_datasink"),
                            name="bold_datasink")

    anat_datasink = pe.Node(DataSink(base_directory=output_folder,
                                     container="anat_datasink"),
                            name="anat_datasink")

    transforms_datasink = pe.Node(DataSink(base_directory=output_folder,
                                           container="transforms_datasink"),
                                  name="transforms_datasink")

    motion_datasink = pe.Node(DataSink(base_directory=output_folder,
                                          container="motion_datasink"),
                                 name="motion_datasink")

    from nipype import logging
    log = logging.getLogger('nipype.workflow')
    import bids
    bids.config.set_option('extension_initial_dot', True)
    try:
        layout = bids.layout.BIDSLayout(data_dir_path, validate=True)
    except Exception as e:
        log.warning(f"The BIDS compliance failed: {e} \n\nRABIES will run anyway; double-check that the right files were picked up for processing.\n")
        layout = bids.layout.BIDSLayout(data_dir_path, validate=False)

    split_name, scan_info, run_iter, structural_scan_list, bold_scan_list = prep_bids_iter(
        layout, opts.bids_filter, opts.bold_only, inclusion_list=opts.inclusion_ids, exclusion_list=opts.exclusion_ids)
    '''***details on outputs from prep_bids_iter:
    split_name: a list of strings, providing a sensible name to distinguish each iterable, 
        and also necessary to link up the run iterables with a specific session later.
    scan_info: a list of dictionary including the subject ID and session # for a given 
        iterable from split_name
    run_iter: a list of dictionary, where the keys correspond to a session split from 
        split_name, and the value is a list of runs for that split. This manages iterables
        for runs.
    structural_scan_list: list of structural file names used for commonspace registration; used for resample_template and managing # threads
    bold_scan_list: list of functional file names; used for resample_template and managing # threads
    '''
    number_structural_scans = len(structural_scan_list)
    number_functional_scans = len(bold_scan_list)

    if opts.num_ITK_threads=='optimal': # override previous assignment from run_main(): an optimal thread number can be selected based on # of scans
        num_ITK_threads=max(int(opts.local_threads/number_functional_scans),1)
        log.info(f"An optimal number of {num_ITK_threads} ITK threads are allocated for {number_functional_scans} functional scans.")
        os.environ['RABIES_ITK_NUM_THREADS']=str(num_ITK_threads)
    else:
        num_ITK_threads = int(os.environ['RABIES_ITK_NUM_THREADS'])

    # calculate the number of threads used for modelbuild
    # the number cannot exceed --local_threads, but should use number_structural_scans*num_ITK_threads at least
    nthreads_modelbuild_anat = min(opts.local_threads, number_structural_scans*num_ITK_threads)

    # setting up all iterables
    main_split = pe.Node(niu.IdentityInterface(fields=['split_name', 'scan_info']),
                         name="main_split")
    main_split.iterables = [('split_name', split_name),
                            ('scan_info', scan_info)]
    main_split.synchronize = True

    '''
    First execute a workflow that finds and formats inputs. 
    This workflow includes a first split and merge of iterables, which is necessary to conduct separately to avoid
    downstream errors with iterables.
    '''
    prep_input_wf = init_prep_input_wf(data_dir_path, split_name, scan_info, run_iter, opts, name='prep_input_wf')

    '''
    MATCHING INPUTS TO MAIN ITERABLES
    '''
    format_bold_node = pe.Node(Function(input_names=['scan_info', 'run', 'joined_scan_info', 'joined_run', 'joined_file_list'],
                                                output_names=['selected_file'],
                                                function=match_iterables),
                                        name='format_bold')
    
    input_bold_node = pe.Node(Function(input_names=['scan_info', 'run', 'joined_scan_info', 'joined_run', 'joined_file_list'],
                                                output_names=['selected_file'],
                                                function=match_iterables),
                                        name='input_bold')

    workflow.connect([
        (prep_input_wf, format_bold_node, [
            ("outputnode.prep_bold_list", "joined_file_list"),
            ("outputnode.joined_scan_info", "joined_scan_info"),
            ]),
        (main_split, format_bold_node, [
            ("scan_info", "scan_info"),
            ]),
        (prep_input_wf, input_bold_node, [
            ("outputnode.input_bold_list", "joined_file_list"),
            ("outputnode.joined_scan_info", "joined_scan_info"),
            ]),
        (main_split, input_bold_node, [
            ("scan_info", "scan_info"),
            ]),
        ])
    
    if opts.bold_only:
        format_bold_node.inputs.run = None
        input_bold_node.inputs.run = None
        format_bold_node.inputs.joined_run = None
        input_bold_node.inputs.joined_run = None
    else:
        run_split = pe.Node(niu.IdentityInterface(fields=['run', 'split_name']),
                            name="run_split")
        run_split.itersource = ('main_split', 'split_name')
        run_split.iterables = [('run', run_iter)]

        workflow.connect([
            (main_split, run_split, [
                ("split_name", "split_name"),
                ]),
            (prep_input_wf, format_bold_node, [
                ("outputnode.joined_run", "joined_run"),
                ]),
            (run_split, format_bold_node, [
                ("run", "run"),
                ]),
            (prep_input_wf, input_bold_node, [
                ("outputnode.joined_run", "joined_run"),
                ]),
            (run_split, input_bold_node, [
                ("run", "run"),
                ]),
            ])

        format_anat_node = pe.Node(Function(input_names=['scan_info', 'run', 'joined_scan_info', 'joined_run', 'joined_file_list'],
                                                    output_names=['selected_file'],
                                                    function=match_iterables),
                                            name='format_anat')
        format_anat_node.inputs.run = None
        format_anat_node.inputs.joined_run = None
        
        input_anat_node = pe.Node(Function(input_names=['scan_info', 'run', 'joined_scan_info', 'joined_run', 'joined_file_list'],
                                                    output_names=['selected_file'],
                                                    function=match_iterables),
                                            name='input_anat')
        input_anat_node.inputs.run = None
        input_anat_node.inputs.joined_run = None
        
        workflow.connect([
            (prep_input_wf, format_anat_node, [
                ("outputnode.prep_anat_list", "joined_file_list"),
                ("outputnode.anat_joined_scan_info", "joined_scan_info"),
                ]),
            (main_split, format_anat_node, [
                ("scan_info", "scan_info"),
                ]),
            (prep_input_wf, input_anat_node, [
                ("outputnode.input_anat_list", "joined_file_list"),
                ("outputnode.anat_joined_scan_info", "joined_scan_info"),
                ]),
            (main_split, input_anat_node, [
                ("scan_info", "scan_info"),
                ]),
            ])


    if opts.inherit_unbiased_template=='none':
        inherit_unbiased=False
        # Resample the anatomical template according to the resolution of the provided input data
        resample_template_node = pe.Node(Function(input_names=['structural_scan_list', 'bold_scan_list', 'template_file', 
                                                               'mask_file', 'anatomical_resampling', 'commonspace_resampling', 
                                                               'rabies_data_type', 'bold_only'],
                                                output_names=[
                                                    'registration_template', 'registration_mask', 'commonspace_template'],
                                                function=resample_template),
                                        name='resample_template', mem_gb=1*opts.scale_min_memory)
        resample_template_node.inputs.template_file = opts.anat_template
        resample_template_node.inputs.mask_file = opts.brain_mask
        resample_template_node.inputs.anatomical_resampling = opts.anatomical_resampling
        resample_template_node.inputs.commonspace_resampling = opts.commonspace_resampling
        resample_template_node.inputs.rabies_data_type = opts.data_type
        resample_template_node.inputs.bold_only = opts.bold_only

        workflow.connect([
            (prep_input_wf, resample_template_node, [
                ("outputnode.prep_anat_list", "structural_scan_list"),
                ("outputnode.prep_bold_list", "bold_scan_list"),
                ]),
            ])

    else: # inherit the atlas files from previous run
        inherit_unbiased=True
        opts, inherit_dict = inherit_unbiased_files(opts.inherit_unbiased_template, opts)
        resample_template_node = pe.Node(niu.IdentityInterface(fields=['registration_template', 'registration_mask', 'commonspace_template']),
                                            name="resample_template")
        resample_template_node.inputs.registration_template = inherit_dict['registration_template']
        resample_template_node.inputs.registration_mask = inherit_dict['registration_mask']
        resample_template_node.inputs.commonspace_template = inherit_dict['commonspace_template']


    EPI_target_buffer = pe.Node(niu.IdentityInterface(fields=['EPI_template', 'EPI_mask']),
                                        name="EPI_target_buffer")

    commonspace_reg_wf = init_commonspace_reg_wf(opts=opts, commonspace_reg_opts=opts.commonspace_reg, inherit_unbiased=inherit_unbiased,
                                                 output_folder=output_folder, transforms_datasink=transforms_datasink, nthreads_modelbuild=nthreads_modelbuild_anat, output_datasinks=True, 
                                                 joinsource_list=['main_split'], name='commonspace_reg_wf')
    if inherit_unbiased:
        commonspace_reg_wf.inputs.inherit_unbiased_inputnode.unbiased_template = inherit_dict['unbiased_template']
        commonspace_reg_wf.inputs.inherit_unbiased_inputnode.unbiased_mask = inherit_dict['unbiased_mask']
        commonspace_reg_wf.inputs.inherit_unbiased_inputnode.unbiased_to_atlas_affine = inherit_dict['unbiased_to_atlas_affine']
        commonspace_reg_wf.inputs.inherit_unbiased_inputnode.unbiased_to_atlas_warp = inherit_dict['unbiased_to_atlas_warp']
        commonspace_reg_wf.inputs.inherit_unbiased_inputnode.unbiased_to_atlas_inverse_warp = inherit_dict['unbiased_to_atlas_inverse_warp']
        commonspace_reg_wf.inputs.inherit_unbiased_inputnode.warped_unbiased = inherit_dict['warped_unbiased']

    bold_main_wf = init_bold_main_wf(opts=opts, output_folder=output_folder, number_functional_scans=number_functional_scans)

    # organizing visual QC outputs
    template_diagnosis = pe.Node(Function(input_names=['anat_template', 'opts', 'out_dir', 'figure_format'],
                                       function=preprocess_visual_QC.template_info),
                              name='template_info')
    template_diagnosis.inputs.opts = opts
    template_diagnosis.inputs.out_dir = output_folder+'/preprocess_QC_report/template_files/'
    template_diagnosis.inputs.figure_format = opts.figure_format

    bold_inho_cor_diagnosis = pe.Node(Function(input_names=['raw_img','init_denoise','warped_mask','final_denoise', 'name_source', 'out_dir', 'figure_format'],
                                       function=preprocess_visual_QC.inho_cor_diagnosis),
                              name='bold_inho_cor_diagnosis')
    bold_inho_cor_diagnosis.inputs.out_dir = output_folder+'/preprocess_QC_report/bold_inho_cor/'
    bold_inho_cor_diagnosis.inputs.figure_format = opts.figure_format

    temporal_diagnosis = pe.Node(Function(input_names=['bold_file', 'motion_params_csv', 'FD_csv', 'rabies_data_type', 'name_source', 'out_dir', 'figure_format'],
                                          output_names=[
                                            'std_filename', 'tSNR_filename'],
                                       function=preprocess_visual_QC.temporal_features),
                              name='temporal_features')
    temporal_diagnosis.inputs.out_dir = output_folder+'/preprocess_QC_report/temporal_features/'
    temporal_diagnosis.inputs.rabies_data_type = opts.data_type
    temporal_diagnosis.inputs.figure_format = opts.figure_format

    # MAIN WORKFLOW STRUCTURE #######################################################
    workflow.connect([
        (format_bold_node, bold_main_wf, [
            ("selected_file", "inputnode.bold"),
            ]),
        (resample_template_node, template_diagnosis, [
            ("registration_template", "anat_template"),
            ]),
        (resample_template_node, commonspace_reg_wf, [
            ("registration_template", "template_inputnode.template_anat"),
            ("registration_mask", "template_inputnode.template_mask"),
            ]),
        (resample_template_node, bold_main_wf, [
            ("commonspace_template", "inputnode.commonspace_ref"),
            ]),
        (resample_template_node, outputnode, [
            ("commonspace_template", "commonspace_resampled_template"),
            ]),
        (commonspace_reg_wf, bold_main_wf, [
            ("outputnode.anat_to_commonspace_transform_list", "inputnode.anat_to_commonspace_transform_list"),
            ("outputnode.anat_to_commonspace_inverse_list", "inputnode.anat_to_commonspace_inverse_list"),
            ("outputnode.commonspace_to_anat_transform_list", "inputnode.commonspace_to_anat_transform_list"),
            ("outputnode.commonspace_to_anat_inverse_list", "inputnode.commonspace_to_anat_inverse_list"),
            ]),
        (bold_main_wf, outputnode, [
            ("outputnode.bold_ref", "initial_bold_ref"),
            ("outputnode.corrected_EPI", "inho_cor_bold"),
            ("outputnode.native_brain_mask", "native_brain_mask"),
            ("outputnode.native_WM_mask", "native_WM_mask"),
            ("outputnode.native_CSF_mask", "native_CSF_mask"),
            ("outputnode.native_vascular_mask", "native_vascular_mask"),
            ("outputnode.motion_params_csv", "motion_params_csv"),
            ("outputnode.FD_voxelwise", "FD_voxelwise"),
            ("outputnode.pos_voxelwise", "pos_voxelwise"),
            ("outputnode.FD_csv", "FD_csv"),
            ('outputnode.bold_to_anat_affine', 'bold_to_anat_affine'),
            ('outputnode.bold_to_anat_warp', 'bold_to_anat_warp'),
            ('outputnode.bold_to_anat_inverse_warp', 'bold_to_anat_inverse_warp'),
            ("outputnode.output_warped_bold", "inho_cor_bold_warped2anat"),
            ("outputnode.native_bold", "native_bold"),
            ("outputnode.native_bold_ref", "native_bold_ref"),
            ("outputnode.commonspace_bold", "commonspace_bold"),
            ("outputnode.commonspace_mask", "commonspace_mask"),
            ("outputnode.commonspace_WM_mask", "commonspace_WM_mask"),
            ("outputnode.commonspace_CSF_mask", "commonspace_CSF_mask"),
            ("outputnode.commonspace_vascular_mask", "commonspace_vascular_mask"),
            ("outputnode.boldspace_brain_mask", "boldspace_brain_mask"),
            ]),
        (bold_main_wf, bold_inho_cor_diagnosis, [
            ("outputnode.bold_ref", "raw_img"),
            ("outputnode.init_denoise", "init_denoise"),
            ("outputnode.corrected_EPI", "final_denoise"),
            ("outputnode.denoise_mask", "warped_mask"),
            ]),
        (input_bold_node, bold_inho_cor_diagnosis,
         [("selected_file", "name_source")]),
        (bold_main_wf, temporal_diagnosis, [
            ("outputnode.motion_params_csv", "motion_params_csv"),
            ("outputnode.FD_csv", "FD_csv"),
            ]),
        (input_bold_node, temporal_diagnosis,
         [("selected_file", "name_source")]),
        (temporal_diagnosis, outputnode, [
            ("tSNR_filename", "tSNR_filename"),
            ("std_filename", "std_filename"),
            ]),
        ])
    
    if opts.generate_commonspace: 
        workflow.connect([
            (bold_main_wf, temporal_diagnosis, [
                ("outputnode.commonspace_bold", "bold_file"),
                ]),
            ])
    else: 
        workflow.connect([
            (bold_main_wf, temporal_diagnosis, [
                ("outputnode.native_bold", "bold_file"),
                ]),
            ])

    if not opts.bold_only:
        # setting anat preprocessing nodes
        anat_inho_cor_wf = init_inho_correction_wf(opts=opts, image_type='structural', output_folder=output_folder, nthreads_modelbuild=nthreads_modelbuild_anat, name="anat_inho_cor_wf")

        workflow.connect([
            (format_anat_node, anat_inho_cor_wf, [
                ("selected_file", "inputnode.target_img"),
                ("selected_file", "inputnode.name_source"),
                ]),
            (resample_template_node, anat_inho_cor_wf, [
                ("registration_template", "inputnode.anat_ref"),
                ("registration_mask", "inputnode.anat_mask"),
                ]),
            (resample_template_node, anat_inho_cor_wf, [
                ("registration_template", "template_inputnode.template_anat"),
                ("registration_mask", "template_inputnode.template_mask"),
                ]),
            (anat_inho_cor_wf, bold_main_wf, [
                ("outputnode.corrected", "inputnode.coreg_anat"),
                ]),
            (commonspace_reg_wf, bold_main_wf, [
                ("outputnode.anatspace_mask", "inputnode.coreg_mask"),
                ("outputnode.unbiased_template", "template_inputnode.template_anat"),
                ("outputnode.unbiased_mask", "template_inputnode.template_mask"),
                ]),
            (anat_inho_cor_wf, commonspace_reg_wf, [
                ("outputnode.corrected", "inputnode.moving_image"),
                ("outputnode.denoise_mask", "inputnode.moving_mask"),
                ]),
            (anat_inho_cor_wf, EPI_target_buffer, [
                ("outputnode.corrected", "EPI_template"),
                ]),
            (commonspace_reg_wf, EPI_target_buffer, [
                ("outputnode.anatspace_mask", 'EPI_mask'),
                ]),
            (EPI_target_buffer, bold_main_wf, [
                ("EPI_template", "inputnode.inho_cor_anat"),
                ("EPI_mask", "inputnode.inho_cor_mask"),
                ]),
            ])

        if not opts.anat_inho_cor['method']=='disable':
            anat_inho_cor_diagnosis = pe.Node(Function(input_names=['raw_img','init_denoise','warped_mask','final_denoise', 'name_source', 'out_dir', 'figure_format'],
                                               function=preprocess_visual_QC.inho_cor_diagnosis),
                                      name='anat_inho_cor_diagnosis')
            anat_inho_cor_diagnosis.inputs.out_dir = output_folder+'/preprocess_QC_report/anat_inho_cor/'
            anat_inho_cor_diagnosis.inputs.figure_format = opts.figure_format

            workflow.connect([
                (format_anat_node, anat_inho_cor_diagnosis, [
                    ("selected_file", "raw_img"),
                    ]),
                (input_anat_node, anat_inho_cor_diagnosis, [
                    ("selected_file", "name_source"),
                    ]),
                (anat_inho_cor_wf, anat_inho_cor_diagnosis, [
                    ("outputnode.init_denoise", "init_denoise"),
                    ("outputnode.corrected", "final_denoise"),
                    ("outputnode.denoise_mask", "warped_mask"),
                    ]),
                ])

    else:
        inho_cor_bold_main_wf = init_bold_main_wf(
            output_folder=output_folder, number_functional_scans=number_functional_scans, inho_cor_only=True, name='inho_cor_bold_main_wf', opts=opts)

        workflow.connect([
            (resample_template_node, inho_cor_bold_main_wf, [
                ("registration_template", "template_inputnode.template_anat"),
                ("registration_mask", "template_inputnode.template_mask"),
                ]),
            (format_bold_node, inho_cor_bold_main_wf, [
                ("selected_file", "inputnode.bold"),
                ]),
            (EPI_target_buffer, inho_cor_bold_main_wf, [
                ("EPI_template", "inputnode.inho_cor_anat"),
                ("EPI_mask", "inputnode.inho_cor_mask"),
                ]),
            (inho_cor_bold_main_wf, bold_main_wf, [
                ("transitionnode.bold_file", "transitionnode.bold_file"),
                ("transitionnode.isotropic_bold_file", "transitionnode.isotropic_bold_file"),
                ("transitionnode.bold_ref", "transitionnode.bold_ref"),
                ("transitionnode.init_denoise", "transitionnode.init_denoise"),
                ("transitionnode.denoise_mask", "transitionnode.denoise_mask"),
                ("transitionnode.corrected_EPI", "transitionnode.corrected_EPI"),
                ("transitionnode.log_bold", "transitionnode.log_bold"),
                ]),
            (inho_cor_bold_main_wf, commonspace_reg_wf, [
                ("transitionnode.corrected_EPI", "inputnode.moving_image"),
                ("transitionnode.denoise_mask", "inputnode.moving_mask"),
                ]),
            (resample_template_node, EPI_target_buffer, [
                ("registration_template", "EPI_template"),
                ("registration_mask", "EPI_mask"),
                ]),
            ])

    if not opts.bold_only:
        PlotOverlap_EPI2Anat_node = pe.Node(
            preprocess_visual_QC.PlotOverlap(), name='PlotOverlap_EPI2Anat')
        PlotOverlap_EPI2Anat_node.inputs.out_dir = output_folder+'/preprocess_QC_report/EPI2Anat'
        workflow.connect([
            (input_bold_node, PlotOverlap_EPI2Anat_node,
             [("selected_file", "name_source")]),
            (anat_inho_cor_wf, PlotOverlap_EPI2Anat_node,
             [("outputnode.corrected", "fixed")]),
            (outputnode, PlotOverlap_EPI2Anat_node, [
                ("inho_cor_bold_warped2anat", "moving"),  # warped EPI to anat
                ]),
            ])

    # fill the datasinks
    workflow.connect([
        (input_bold_node, bold_datasink, [
            ("selected_file", "input_bold"),
            ]),
        (outputnode, motion_datasink, [
            ("motion_params_csv", "motion_params_csv"),  # confounds file
            ("FD_voxelwise", "FD_voxelwise"),
            ("pos_voxelwise", "pos_voxelwise"),
            ("FD_csv", "FD_csv"),
            ]),
        (bold_main_wf, motion_datasink, [
            ('outputnode.hmc_qc_figure', 'hmc_qc_figure'),
            ('outputnode.hmc_qc_csv', 'hmc_qc_csv'),
            ('outputnode.hmc_qc_video', 'hmc_qc_video'),
            ]),
        (outputnode, bold_datasink, [
            ("initial_bold_ref", "initial_bold_ref"),  # inspect initial bold ref
            ("inho_cor_bold", "inho_cor_bold"),  # inspect bias correction
            ("native_brain_mask", "native_brain_mask"),  # get the EPI labels
            ("native_WM_mask", "native_WM_mask"),  # get the EPI labels
            ("native_CSF_mask", "native_CSF_mask"),  # get the EPI labels
            ("native_vascular_mask", "native_vascular_mask"),  # get the EPI labels
            # warped EPI to anat
            ("inho_cor_bold_warped2anat", "inho_cor_bold_warped2anat"),
            # resampled EPI after motion realignment and SDC
            ("native_bold", "native_bold"),
            # resampled EPI after motion realignment and SDC
            ("native_bold_ref", "native_bold_ref"),
            # resampled EPI after motion realignment and SDC
            ("commonspace_bold", "commonspace_bold"),
            ("commonspace_mask", "commonspace_mask"),
            ("commonspace_WM_mask", "commonspace_WM_mask"),
            ("commonspace_CSF_mask", "commonspace_CSF_mask"),
            ("commonspace_vascular_mask", "commonspace_vascular_mask"),
            ("tSNR_filename", "tSNR_map_preprocess"),
            ("std_filename", "std_map_preprocess"),
            ("commonspace_resampled_template", "commonspace_resampled_template"),
            ("boldspace_brain_mask", "boldspace_brain_mask"),
            ]),
        ])

    if not opts.bold_only:
        workflow.connect([
            (anat_inho_cor_wf, anat_datasink, [
                ("outputnode.corrected", "anat_preproc"),
                ]),
            (outputnode, transforms_datasink, [
                ('bold_to_anat_affine', 'bold_to_anat_affine'),
                ('bold_to_anat_warp', 'bold_to_anat_warp'),
                ('bold_to_anat_inverse_warp', 'bold_to_anat_inverse_warp'),
                ]),
            ])

    return workflow



import os
from nipype.interfaces import utility as niu
import nipype.pipeline.engine as pe  # pypeline engine
from nipype.interfaces.utility import Function
from nipype.interfaces.base import (
    traits, TraitedSpec, BaseInterfaceInputSpec,
    File, BaseInterface
)
from nipype.interfaces.io import DataSink
from rabies.utils import run_command, flatten_list, ResampleMask
from .registration import run_antsRegistration
from .preprocess_visual_QC import PlotOverlap,template_masking

def init_commonspace_reg_wf(opts, commonspace_reg_opts, inherit_unbiased, output_folder, transforms_datasink, nthreads_modelbuild, output_datasinks, joinsource_list, name='commonspace_reg_wf'):
    # commonspace_wf_head_start
    """
    This workflow handles the alignment of all MRI sessions to a common space. This is conducted first by generating
    a dataset-specific unbiased template from the input structural images, thereby aligning the different MRI 
    sessions. Through a set of iterations, images are registered to a consensus average generated from the overlap 
    of all scans at the previous iteration. Registrations are increasingly stringent, executing 2 iterations each 
    for a rigid, then affine and finally non-linear template generation. The final iteration provides the individual 
    transforms to align each scan to the unbiased template. This template generation process creates a robust target 
    for the alignment of MRI sessions sharing the same acquisition properties, which will minimize registration 
    inconsistencies between sessions, as opposed to the direct registration to an external template. The algorithm is 
    implemented in https://github.com/CoBrALab/optimized_antsMultivariateTemplateConstruction.
    After generating the unbiased template, the template itself is registered with a non-linear registration to the
    reference atlas in common space, providing transforms to common space and the associated brain parcellations.


    References:
        Avants, B. B., Tustison, N. J., Song, G., Cook, P. A., Klein, A., & Gee, J. C. (2011). A reproducible evaluation 
        of ANTs similarity metric performance in brain image registration. NeuroImage, 54(3), 2033–2044.

    Command line interface parameters:
        --commonspace_reg COMMONSPACE_REG
                                Specify registration options for the commonspace registration.
                                * masking: Combine masks derived from the inhomogeneity correction step to support 
                                registration during the generation of the unbiased template, and then during template 
                                registration.
                                *** Specify 'true' or 'false'. 
                                * brain_extraction: conducts brain extraction prior to template registration based on the 
                                combined masks from inhomogeneity correction. This will enhance brain edge-matching, but 
                                requires good quality masks. This should be selected along the 'masking' option.
                                *** Specify 'true' or 'false'. 
                                * keep_mask_after_extract: If using brain_extraction, use the mask to compute the registration metric
                                within the mask only. Choose to prevent stretching of the images beyond the limit of the brain mask
                                (e.g. if the moving and target images don't have the same brain coverage).
                                *** Specify 'true' or 'false'.
                                * template_registration: Specify a registration script for the alignment of the 
                                dataset-generated unbiased template to the commonspace atlas.
                                *** Rigid: conducts only rigid registration.
                                *** Affine: conducts Rigid then Affine registration.
                                *** SyN: conducts Rigid, Affine then non-linear registration.
                                *** no_reg: skip registration.
                                * fast_commonspace: Skip the generation of a dataset-generated unbiased template, and 
                                instead, register each scan independently directly onto the commonspace atlas, using the 
                                template_registration. This option can be faster, but may decrease the quality of 
                                alignment between subjects. 
                                *** Specify 'true' or 'false'. 
                                (default: masking=false,brain_extraction=false,keep_mask_after_extract=false,template_registration=SyN,fast_commonspace=false)
                        
    Workflow:
        parameters
            opts: command line interface parameters (technically all the parser args, including winsorization and brain_extraction etc are included here)
            commonspace_reg_opts: the opts that are specific to this run of commonspace_reg (there can be different parameters, for instance for the robust_inho_cor workflows)
            output_folder: specify a folder to execute the workflow and store important outputs
            transforms_datasink: datasink node where the transforms are stored
            nthreads_modelbuild: set the maximum number of parallel threads to launch
            output_datasinks: whether to generate a datasink from the outputs of the workflow
            joinsource_list: names for the iterable nodes to join before unbiased template generation

        inputs
            moving_image_list: list of files corresponding to the images from different MRI sessions
            moving_mask_list: mask files overlapping with the moving images, inherited from the inhomogeneity 
                correction step. These masks are used for --commonspace_masking and --brain_extraction
            template_anat: the target structural template to register the unbiased template
            template_mask: the brain mask of the structural template

        outputs
            unbiased_template: the generated unbiased template
            unbiased_mask: brain mask resampled over the unbiased template
            anatspace_mask: the atlas brain mask resampled to an associated MRI session in anat space
            to_atlas_affine: affine transform for registration to the atlas
            to_atlas_warp: non-linear transform for registration to the atlas
            to_atlas_inverse_warp: inverse of the non-linear transform for registration to the atlas
            anat_to_unbiased_affine: affine transform from anat space to the unbiased template
            anat_to_unbiased_warp: non-linear transform from anat space to the unbiased template
            anat_to_unbiased_inverse_warp: inverse of the non-linear transform from anat space to the unbiased template
            anat_to_commonspace_transform_list: ordered list of the transforms to apply to move from the
                anat space to the common space
            anat_to_commonspace_inverse_list: list defining whether the inverse of affine transforms should
                be applied for anat_to_commonspace_transform_list
            commonspace_to_anat_transform_list: ordered list of the transforms to apply to move from the
                common space to the anat space
            commonspace_to_anat_inverse_list: list defining whether the inverse of affine transforms should
                be applied for commonspace_to_anat_transform_list
    """
    # commonspace_wf_head_end

    # this iterable node inherits the iterations from the main_wf, generating an iteration for each session to recover its outputs from commonspace registration
    inputnode = pe.Node(niu.IdentityInterface(fields=['moving_image', 'moving_mask']),
                                        name="inputnode")

    template_inputnode = pe.Node(niu.IdentityInterface(fields=['template_anat', 'template_mask']),
                                        name="template_inputnode")
    outputnode = pe.Node(niu.IdentityInterface(fields=['unbiased_template', 'unbiased_mask', 'anatspace_mask', 'to_atlas_affine', 'to_atlas_warp', 'to_atlas_inverse_warp',
                                                       'anat_to_unbiased_affine', 'anat_to_unbiased_warp', 'anat_to_unbiased_inverse_warp',
                                                       'anat_to_commonspace_transform_list','anat_to_commonspace_inverse_list',
                                                       'commonspace_to_anat_transform_list','commonspace_to_anat_inverse_list']),
                                        name="outputnode")
    workflow = pe.Workflow(name=name)

    modelbuild_stages = commonspace_reg_opts['stages']
    commonspace_masking = commonspace_reg_opts['masking']
    brain_extraction = commonspace_reg_opts['brain_extraction']
    keep_mask_after_extract = commonspace_reg_opts['keep_mask_after_extract']
    template_reg = commonspace_reg_opts['template_registration']
    fast_commonspace = commonspace_reg_opts['fast_commonspace']
    winsorize_lower_bound = commonspace_reg_opts['winsorize_lower_bound']
    winsorize_upper_bound = commonspace_reg_opts['winsorize_upper_bound']

    if fast_commonspace or inherit_unbiased:
        # without modelbuild, the inputs iterables are not merged
        source_join_common_reg = pe.Node(niu.IdentityInterface(fields=['file_list0', 'file_list1']),
                                            name="fast_commonreg_buffer")
        merged_join_common_reg = source_join_common_reg
    else:
        workflow, source_join_common_reg, merged_join_common_reg = join_iterables(workflow=workflow, joinsource_list=joinsource_list, node_prefix='commonspace_reg', num_inputs=2)

    if inherit_unbiased:
        # Identity node to relate inherited files to existing workflow
        atlas_reg = pe.Node(niu.IdentityInterface(fields=['moving_image', 'moving_mask', 'fixed_image', 'fixed_mask', 'affine', 'warp',
                                                    'inverse_warp', 'warped_image']),
                                            name="atlas_reg_inherited")
    else:
        atlas_reg = pe.Node(Function(input_names=['reg_method', 'brain_extraction', 'keep_mask_after_extract', 'moving_image', 'moving_mask', 'fixed_image', 'fixed_mask', 'winsorize_lower_bound', 'winsorize_upper_bound','rabies_data_type'],
                                        output_names=['affine', 'warp',
                                                    'inverse_warp', 'warped_image'],
                                        function=run_antsRegistration),
                            name='atlas_reg', mem_gb=2*opts.scale_min_memory, n_procs=int(os.environ['RABIES_ITK_NUM_THREADS']))

        # don't use brain extraction without a moving mask
        if brain_extraction:
            if not commonspace_masking:
                brain_extraction=False

        atlas_reg.inputs.winsorize_lower_bound = winsorize_lower_bound
        atlas_reg.inputs.winsorize_upper_bound = winsorize_upper_bound
        atlas_reg.inputs.brain_extraction = brain_extraction
        atlas_reg.inputs.keep_mask_after_extract = keep_mask_after_extract
        atlas_reg.inputs.rabies_data_type = opts.data_type
        atlas_reg.plugin_args = {
            'qsub_args': f'-pe smp {str(3*opts.min_proc)}', 'overwrite': True}
        atlas_reg.inputs.reg_method = template_reg

    prep_commonspace_transform_node = pe.Node(Function(input_names=['anatspace_ref', 'atlas_mask', 'anat_to_unbiased_affine',
                                                               'anat_to_unbiased_warp','anat_to_unbiased_inverse_warp',
                                                               'to_atlas_affine','to_atlas_warp','to_atlas_inverse_warp',
                                                               'fast_commonspace'],
                                               output_names=[
                                                   'anatspace_mask', 'anat_to_commonspace_transform_list','anat_to_commonspace_inverse_list','commonspace_to_anat_transform_list','commonspace_to_anat_inverse_list'],
                                               function=prep_commonspace_transform),
                                      name='prep_commonspace_transform')
    prep_commonspace_transform_node.inputs.atlas_mask = str(opts.brain_mask)
    prep_commonspace_transform_node.inputs.fast_commonspace = fast_commonspace

    workflow.connect([
        (inputnode, source_join_common_reg, [
            ("moving_image", "file_list0"),
            ("moving_mask", "file_list1"),
            ]),
        (template_inputnode, atlas_reg, [
            ("template_anat", "fixed_image"),
            ("template_mask", "fixed_mask"),
            ]),
        (atlas_reg, prep_commonspace_transform_node, [
            ("affine", "to_atlas_affine"),
            ("warp", "to_atlas_warp"),
            ("inverse_warp", "to_atlas_inverse_warp"),
            ]),
        (atlas_reg, outputnode, [
            ("affine", "to_atlas_affine"),
            ("warp", "to_atlas_warp"),
            ("inverse_warp", "to_atlas_inverse_warp"),
            ]),
        (prep_commonspace_transform_node, outputnode, [
            ('anat_to_commonspace_transform_list', 'anat_to_commonspace_transform_list'),
            ('anat_to_commonspace_inverse_list', 'anat_to_commonspace_inverse_list'),
            ('commonspace_to_anat_transform_list', 'commonspace_to_anat_transform_list'),
            ('commonspace_to_anat_inverse_list', 'commonspace_to_anat_inverse_list'),
            ('anatspace_mask', 'anatspace_mask'),
            ]),
        ])

    if fast_commonspace:
        prep_commonspace_transform_node.inputs.anat_to_unbiased_affine = None
        prep_commonspace_transform_node.inputs.anat_to_unbiased_warp = None
        prep_commonspace_transform_node.inputs.anat_to_unbiased_inverse_warp = None

        PlotOverlap_Anat2Atlas_node = pe.Node(
            PlotOverlap(), name='PlotOverlap_Anat2Atlas')
        PlotOverlap_Anat2Atlas_node.inputs.out_dir = output_folder+f'/preprocess_QC_report/{name}.Anat2Atlas/'

        workflow.connect([
            (merged_join_common_reg, atlas_reg, [
                ("file_list0", "moving_image"),
                ]),
            (merged_join_common_reg, prep_commonspace_transform_node, [
                ("file_list0", "anatspace_ref"),
                ]),
            (inputnode, PlotOverlap_Anat2Atlas_node, [
                ("moving_image", "name_source"),
                ]),
            (template_inputnode, PlotOverlap_Anat2Atlas_node,[
                ("template_anat", "fixed"),
                ]),
            (atlas_reg, PlotOverlap_Anat2Atlas_node, [
                ("warped_image", "moving"),
                ]),
            ])
        if commonspace_masking:
            workflow.connect([
                (merged_join_common_reg, atlas_reg, [
                    ("file_list1", "moving_mask"),
                    ]),
                ])
        if output_datasinks:
            workflow.connect([
                (atlas_reg, transforms_datasink, [
                    ("affine", "anat_to_atlas_affine"),
                    ("warp", "anat_to_atlas_warp"),
                    ("inverse_warp", "anat_to_atlas_inverse_warp"),
                    ]),
                ])

    else:

        if inherit_unbiased:
            # Identity node to relate inherited files to existing workflow
            generate_template = pe.Node(niu.IdentityInterface(fields=['moving_image_list', 'moving_mask_list', 'template_anat', 
                                                                      'unbiased_template', 'unbiased_mask', 'affine_list', 'warp_list', 
                                                                      'inverse_warp_list', 'warped_image_list']),
                                                name="generate_template_inherited")
        else:
            # setup a node to select the proper files associated with a given input scan for commonspace registration
            commonspace_selectfiles = pe.Node(Function(input_names=['filename', 'anat_list', 'affine_list', 'warp_list', 'inverse_warp_list', 'warped_anat_list'],
                                                    output_names=[
                                                        'anatspace_ref', 'warped_anat', 'anat_to_unbiased_affine','anat_to_unbiased_warp','anat_to_unbiased_inverse_warp'],
                                                    function=select_commonspace_outputs),
                                            name='commonspace_selectfiles')

            generate_template_outputs = f'{output_folder}/main_wf/{name}/generate_template'
            # the plugin is set here instead of as a node input to avoid re-running the nipype workflow if plugin is changed 
            os.environ['MODELBUILD_PLUGIN'] = opts.plugin
            # each modelbuild thread will be running a registration, and each registration uses RABIES_ITK_NUM_THREADS threads
            os.environ['MODELBUILD_NCORES'] = str(int(nthreads_modelbuild//int(os.environ['RABIES_ITK_NUM_THREADS'])))
            generate_template = pe.Node(GenerateTemplate(stages=modelbuild_stages, masking=commonspace_masking, output_folder=generate_template_outputs, winsorize_lower_bound = winsorize_lower_bound, winsorize_upper_bound = winsorize_upper_bound,
                                                ),
                                        name='generate_template', n_procs=nthreads_modelbuild, mem_gb=1*nthreads_modelbuild*opts.scale_min_memory)

        PlotOverlap_Anat2Unbiased_node = pe.Node(
            PlotOverlap(), name='PlotOverlap_Anat2Unbiased')
        PlotOverlap_Anat2Unbiased_node.inputs.out_dir = output_folder+f'/preprocess_QC_report/{name}.Anat2Unbiased/'
        PlotOverlap_Unbiased2Atlas_node = pe.Node(
            PlotOverlap(), name='PlotOverlap_Unbiased2Atlas')
        PlotOverlap_Unbiased2Atlas_node.inputs.out_dir = output_folder+f'/preprocess_QC_report/{name}.Unbiased2Atlas'
        PlotOverlap_Unbiased2Atlas_node.inputs.name_source = ''

        def prep_atlas_to_unbiased_transforms(to_atlas_affine,to_atlas_inverse_warp):
            transform_list=[to_atlas_affine,to_atlas_inverse_warp]
            inverse_list=[1,0]
            return transform_list,inverse_list
        atlas_to_unbiased_transforms_node = pe.Node(Function(input_names=['to_atlas_affine','to_atlas_inverse_warp'],
                                                output_names=[
                                                    'transform_list','inverse_list'],
                                                function=prep_atlas_to_unbiased_transforms),
                                        name='atlas_to_unbiased_transforms')

        resample_unbiased_mask_node = pe.Node(ResampleMask(), name='resample_unbiased_mask')
        resample_unbiased_mask_node.inputs.name_suffix = 'unbiased_mask_resampled'

        if brain_extraction and commonspace_masking:
            template_masking_node = pe.Node(Function(input_names=['template', 'mask', 'out_dir', 'figure_format'],
                                            function=template_masking),
                                    name='template_masking')
            template_masking_node.inputs.out_dir = output_folder+f'/preprocess_QC_report/{name}.unbiased_template_masking/'
            template_masking_node.inputs.figure_format = opts.figure_format

            workflow.connect([
                (generate_template, template_masking_node, [
                    ("unbiased_template", "template"),
                    ("unbiased_mask", "mask"),
                    ]),
                ])

        workflow.connect([
            (template_inputnode, generate_template, [
                ("template_anat", "template_anat"),
                ]),
            (merged_join_common_reg, generate_template, [
                ("file_list0", "moving_image_list"),
                ]),
            (template_inputnode, PlotOverlap_Unbiased2Atlas_node,[
                ("template_anat", "fixed"),
                ]),
            (inputnode, PlotOverlap_Anat2Unbiased_node, [
                ("moving_image", "name_source"),
                ]),
            (generate_template, atlas_reg, [
                ("unbiased_template", "moving_image"),
                ]),
            (generate_template, outputnode, [
                ("unbiased_template", "unbiased_template"),
                ]),
            (atlas_reg, atlas_to_unbiased_transforms_node, [
                ("affine", "to_atlas_affine"),
                ("inverse_warp", "to_atlas_inverse_warp"),
                ]),
            (atlas_to_unbiased_transforms_node, resample_unbiased_mask_node, [
                ("transform_list", "transforms"),
                ("inverse_list", "inverses"),
                ]),
            (generate_template, resample_unbiased_mask_node, [
                ("unbiased_template", "ref_file"),
                ]),
            (template_inputnode, resample_unbiased_mask_node, [
                ("template_mask", "mask_file"),
                ("template_mask", "name_source"),
                ]),
            (resample_unbiased_mask_node, outputnode, [
                ("resampled_file", "unbiased_mask"),
                ]),
            (generate_template, PlotOverlap_Anat2Unbiased_node, [
                ("unbiased_template", "fixed"),
                ]),
            (atlas_reg, PlotOverlap_Unbiased2Atlas_node, [
                ("warped_image", "moving"),
                ]),
            ])
        if output_datasinks:
            unbiased_template_datasink = pe.Node(DataSink(base_directory=output_folder,
                                                container="unbiased_template_datasink"),
                                        name="unbiased_template_datasink")

            workflow.connect([
                (outputnode, unbiased_template_datasink, [
                    ("unbiased_template", "unbiased_template"),
                    ]),
                (atlas_reg, unbiased_template_datasink, [
                    ("warped_image", "warped_unbiased_template"),
                    ]),
                ])

        if commonspace_masking:
            workflow.connect([
                (merged_join_common_reg, generate_template, [
                    ("file_list1", "moving_mask_list"),
                    ]),
                (generate_template, atlas_reg, [
                    ("unbiased_mask", "moving_mask"),
                    ]),
                ])

        if output_datasinks:
            workflow.connect([
                (atlas_reg, transforms_datasink, [
                    ("affine", "unbiased_to_atlas_affine"),
                    ("warp", "unbiased_to_atlas_warp"),
                    ("inverse_warp", "unbiased_to_atlas_inverse_warp"),
                    ]),
                ])

        if inherit_unbiased:
            inherit_unbiased_inputnode = pe.Node(niu.IdentityInterface(fields=['unbiased_template', 'unbiased_mask', 'unbiased_to_atlas_affine', 'unbiased_to_atlas_warp',
                                                                            'unbiased_to_atlas_inverse_warp', 'warped_unbiased']),
                                                name="inherit_unbiased_inputnode")

            inherit_unbiased_reg_node = pe.Node(Function(input_names=['reg_method', 'brain_extraction', 'keep_mask_after_extract', 'moving_image', 'moving_mask', 'fixed_image', 'fixed_mask', 'winsorize_lower_bound', 'winsorize_upper_bound','rabies_data_type'],
                                            output_names=['affine', 'warp',
                                                        'inverse_warp', 'warped_image'],
                                            function=run_antsRegistration),
                                name='inherit_unbiased_reg', mem_gb=2*opts.scale_min_memory)
            inherit_unbiased_reg_node.inputs.winsorize_lower_bound = winsorize_lower_bound
            inherit_unbiased_reg_node.inputs.winsorize_upper_bound = winsorize_upper_bound
            inherit_unbiased_reg_node.inputs.reg_method = 'Rigid' # this is the registration modelbuild conducts
            inherit_unbiased_reg_node.inputs.brain_extraction = False # brain extraction is not applied during template building
            inherit_unbiased_reg_node.inputs.keep_mask_after_extract = False # brain extraction is not applied during template building
            inherit_unbiased_reg_node.inputs.rabies_data_type = opts.data_type

            if commonspace_masking: # this parameter should be inherited from previous run
                workflow.connect([
                    (inputnode, inherit_unbiased_reg_node, [
                        ("moving_mask", "moving_mask"),
                        ]),
                    (generate_template, inherit_unbiased_reg_node, [
                        ("unbiased_mask", "fixed_mask"),
                        ]),
                    ])
            else:
                inherit_unbiased_reg_node.inputs.moving_mask = 'NULL'
                inherit_unbiased_reg_node.inputs.fixed_mask = 'NULL'

            workflow.connect([
                (inherit_unbiased_inputnode, generate_template, [
                    ("unbiased_template", "unbiased_template"),
                    ("unbiased_mask", "unbiased_mask"),
                    ]),
                (inherit_unbiased_inputnode, atlas_reg, [
                    ("unbiased_to_atlas_affine", "affine"),
                    ("unbiased_to_atlas_warp", "warp"),
                    ("unbiased_to_atlas_inverse_warp", "inverse_warp"),
                    ("warped_unbiased", "warped_image"),
                    ]),
                (inputnode, inherit_unbiased_reg_node, [
                    ("moving_image", "moving_image"),
                    ]),
                (generate_template, inherit_unbiased_reg_node, [
                    ("unbiased_template", "fixed_image"),
                    ]),
                (inherit_unbiased_reg_node, prep_commonspace_transform_node, [
                    ("affine", "anat_to_unbiased_affine"),
                    ("warp", "anat_to_unbiased_warp"),
                    ("inverse_warp", "anat_to_unbiased_inverse_warp"),
                    ]),
                (inherit_unbiased_reg_node, outputnode, [
                    ("affine", "anat_to_unbiased_affine"),
                    ("warp", "anat_to_unbiased_warp"),
                    ("inverse_warp", "anat_to_unbiased_inverse_warp"),
                    ]),
                (merged_join_common_reg, prep_commonspace_transform_node, [
                    ("file_list0", "anatspace_ref"),
                    ]),
                (inherit_unbiased_reg_node, PlotOverlap_Anat2Unbiased_node, [
                    ("warped_image", "moving"),
                    ]),
                ])
            
        else:
            workflow.connect([
                (merged_join_common_reg, commonspace_selectfiles, [
                    ("file_list0", "anat_list"),
                    ]),
                (inputnode, commonspace_selectfiles, [
                    ("moving_image", "filename"),
                    ]),
                (generate_template, commonspace_selectfiles, [
                    ("affine_list", "affine_list"),
                    ("warp_list", "warp_list"),
                    ("inverse_warp_list", "inverse_warp_list"),
                    ("warped_image_list", "warped_anat_list"),
                    ]),
                (commonspace_selectfiles, prep_commonspace_transform_node, [
                    ("anatspace_ref", "anatspace_ref"),
                    ("anat_to_unbiased_affine", "anat_to_unbiased_affine"),
                    ("anat_to_unbiased_warp", "anat_to_unbiased_warp"),
                    ("anat_to_unbiased_inverse_warp", "anat_to_unbiased_inverse_warp"),
                    ]),
                (commonspace_selectfiles, outputnode, [
                    ("anat_to_unbiased_affine", "anat_to_unbiased_affine"),
                    ("anat_to_unbiased_warp", "anat_to_unbiased_warp"),
                    ("anat_to_unbiased_inverse_warp", "anat_to_unbiased_inverse_warp"),
                    ]),
                (commonspace_selectfiles, PlotOverlap_Anat2Unbiased_node, [
                    ("warped_anat", "moving"),
                    ]),
                ])
            
            if output_datasinks:
                workflow.connect([
                    (commonspace_selectfiles, transforms_datasink, [
                        ("anat_to_unbiased_affine", "anat_to_unbiased_affine"),
                        ("anat_to_unbiased_warp", "anat_to_unbiased_warp"),
                        ("anat_to_unbiased_inverse_warp", "anat_to_unbiased_inverse_warp"),
                        ]),
                    ])


    return workflow


def join_iterables(workflow, joinsource_list, node_prefix, num_inputs=1):

    field_list=[]
    for j in range(num_inputs):
        field_list.append(f'file_list{j}')

    i=0
    for joinsource in joinsource_list:
        joinnode = pe.JoinNode(niu.IdentityInterface(fields=field_list),
                                            name=f"{node_prefix}_{joinsource}_joinnode",
                                            joinsource=joinsource,
                                            joinfield=field_list)
        if i==0:
            source_join = joinnode
        else:
            for field in field_list:
                workflow.connect([
                    (joinnode_prev, joinnode, [
                        (field, field),
                        ]),
                    ])

        joinnode_prev = joinnode
        i+=1

    merged_join = joinnode

    return workflow, source_join, merged_join


def select_commonspace_outputs(filename, anat_list, affine_list, warp_list, inverse_warp_list, warped_anat_list):
    from rabies.preprocess_pkg.commonspace_reg import select_from_list
    anat_to_unbiased_affine = select_from_list(filename, affine_list)
    anat_to_unbiased_warp = select_from_list(filename, warp_list)
    anat_to_unbiased_inverse_warp = select_from_list(
        filename, inverse_warp_list)
    warped_anat = select_from_list(filename, warped_anat_list)
    anatspace_ref = select_from_list(filename, anat_list)
    return anatspace_ref, warped_anat, anat_to_unbiased_affine, anat_to_unbiased_warp, anat_to_unbiased_inverse_warp


def select_from_list(filename, filelist):
    filelist = flatten_list(filelist)
    import pathlib
    filename_template = pathlib.Path(filename).name.rsplit(".nii")[0]
    selected_file = None
    for file in filelist:
        if filename_template in file:
            if selected_file is None:
                selected_file = file
            else:
                raise ValueError(
                    f"Found duplicates for filename {filename_template}.")

    if selected_file is None:
        raise ValueError(f"No file associated with {filename_template} were found from list {filelist}.")
    else:
        return selected_file


def prep_commonspace_transform(anatspace_ref, atlas_mask, anat_to_unbiased_affine,
                               anat_to_unbiased_warp,anat_to_unbiased_inverse_warp,
                               to_atlas_affine,to_atlas_warp,to_atlas_inverse_warp, fast_commonspace=False):

    if fast_commonspace:
        anat_to_commonspace_transform_list=[to_atlas_warp,to_atlas_affine]
        anat_to_commonspace_inverse_list=[0,0]
        commonspace_to_anat_transform_list=[to_atlas_affine,to_atlas_inverse_warp]
        commonspace_to_anat_inverse_list=[1,0]
    else:
        anat_to_commonspace_transform_list=[to_atlas_warp,to_atlas_affine,anat_to_unbiased_warp,anat_to_unbiased_affine]
        anat_to_commonspace_inverse_list=[0,0,0,0]
        commonspace_to_anat_transform_list=[anat_to_unbiased_affine,anat_to_unbiased_inverse_warp,to_atlas_affine,to_atlas_inverse_warp]
        commonspace_to_anat_inverse_list=[1,0,1,0]

    import os
    import pathlib  # Better path manipulation
    filename_split = pathlib.Path(
        anatspace_ref).name.rsplit(".nii")
    anatspace_mask = os.path.abspath(filename_split[0]+'_mask.nii.gz')
    from rabies.utils import antsApplyTransforms
    # resample the atlas brain mask to anat space
    antsApplyTransforms(transforms = commonspace_to_anat_transform_list, inverses = commonspace_to_anat_inverse_list, 
                       input_image = atlas_mask, ref_image = anatspace_ref, output_filename = anatspace_mask, interpolation='GenericLabel', rabies_data_type=None, clip_negative=False)

    return anatspace_mask, anat_to_commonspace_transform_list,anat_to_commonspace_inverse_list,commonspace_to_anat_transform_list,commonspace_to_anat_inverse_list


class GenerateTemplateInputSpec(BaseInterfaceInputSpec):
    #these input traits are later inherited in the GenerateTemplate class
    moving_image_list = traits.List(exists=True, mandatory=True,
                            desc="List of anatomical images used for commonspace registration.")
    moving_mask_list = traits.List(exists=True,
                            desc="List of masks accompanying each image. (optional)")
    stages = traits.Str(
        exists=True, mandatory=True, desc="Input to --stages parameter of modelbuild.sh")
    masking = traits.Bool(
        desc="Whether to use the masking option.")
    output_folder = traits.Str(
        exists=True, mandatory=True, desc="Path to output folder.")
    template_anat = File(exists=True, mandatory=True,
                         desc="Reference anatomical template to define the target space.")
    winsorize_lower_bound = traits.Float(
        desc="")
    winsorize_upper_bound = traits.Float(
        desc="")

class GenerateTemplateOutputSpec(TraitedSpec):
    unbiased_template = File(
        exists=True, mandatory=True, desc="Output template generated from commonspace registration.")
    unbiased_mask = File(desc="Output mask generated along with the template. (optional)")
    affine_list = traits.List(exists=True, mandatory=True,
                              desc="List of affine transforms from anat to template space.")
    warp_list = traits.List(exists=True, mandatory=True,
                            desc="List of non-linear transforms from anat to template space.")
    inverse_warp_list = traits.List(exists=True, mandatory=True,
                                    desc="List of the inverse non-linear transforms from anat to template space.")
    warped_image_list = traits.List(exists=True, mandatory=True,
                                   desc="List of anatomical images warped to template space..")


class GenerateTemplate(BaseInterface):
    """

    """

    input_spec = GenerateTemplateInputSpec
    output_spec = GenerateTemplateOutputSpec

    def _run_interface(self, runtime):
        import os
        import pandas as pd
        import pathlib
        import multiprocessing
        from nipype import logging
        log = logging.getLogger('nipype.workflow')

        cwd = os.getcwd()
        template_folder = self.inputs.output_folder
        winsorize_lower_bound = self.inputs.winsorize_lower_bound
        winsorize_upper_bound = self.inputs.winsorize_upper_bound
        stages = self.inputs.stages

        # change the - split for a comma ,
        stages = ','.join(stages.split('-'))

        for stage in stages.split(','):
            if stage not in ['rigid', 'similarity', 'affine', 'nlin']:
                raise ValueError(f"{stage} is not a proper modelbuild.sh stage - it must be one of 'rigid', 'similarity', 'affine', 'nlin'. Review your --commonspace_reg parameter.")

        command = f'mkdir -p {template_folder}'
        rc,c_out = run_command(command)

        merged = flatten_list(list(self.inputs.moving_image_list))
        # create a csv file of the input image list
        csv_path = cwd+'/commonspace_input_files.csv'
        df = pd.DataFrame(data=merged)
        df.to_csv(csv_path, header=False, sep=',', index=False)

        if self.inputs.masking:
            merged_masks = flatten_list(list(self.inputs.moving_mask_list))
            for mask in merged_masks:
                if 'NULL' in mask:
                    merged_masks = ['NULL'] # there should be no NULL mask
            if not (merged_masks[0]=='NULL'):
                mask_csv_path = cwd+'/commonspace_input_masks.csv'
                df = pd.DataFrame(data=merged_masks)
                df.to_csv(mask_csv_path, header=False, sep=',', index=False)
                masks = f'--masks {mask_csv_path}'
            else:
                log.info(f"Some 'NULL' input mask was found within moving_mask_list. No moving mask will be used by modelbuild.sh")
        else:
            merged_masks = ['NULL']
            masks=''
            unbiased_mask='NULL'

        if len(merged) == 1:
            log.info("Only a single scan was provided as input for commonspace registration. Commonspace registration "
                  "won't be run, and the output template will be the input scan.")

            # create an identity transform as a surrogate for the commonspace transforms
            import SimpleITK as sitk
            dimension = 3
            identity = sitk.Transform(dimension, sitk.sitkIdentity)

            file = merged[0]
            mask = merged_masks[0]
            filename_template = pathlib.Path(file).name.rsplit(".nii")[0]
            transform_file = template_folder+filename_template+'_identity.mat'
            sitk.WriteTransform(identity, transform_file)

            setattr(self, 'unbiased_template', file)
            setattr(self, 'unbiased_mask', mask)
            setattr(self, 'affine_list', [transform_file])
            setattr(self, 'warp_list', [transform_file])
            setattr(self, 'inverse_warp_list', [transform_file])
            setattr(self, 'warped_image_list', [file])

            return runtime


        # convert nipype plugin spec to match QBATCH
        plugin = os.environ['MODELBUILD_PLUGIN']
        if plugin=='MultiProc' or plugin=='Linear':
            cluster_type='local'
            num_threads = os.environ['MODELBUILD_NCORES']
        elif plugin=='SGE' or plugin=='SGEGraph':
            cluster_type='sge'
            num_threads = 1
        elif plugin=='PBS':
            cluster_type='PBS'
            num_threads = 1
        elif plugin=='SLURM' or plugin=='SLURMGraph':
            cluster_type='slurm'
            num_threads = 1
        else:
            raise ValueError("Plugin option must correspond to one of 'local', 'sge', 'pbs' or 'slurm'")

        # when the initial target template has the filename of a previous modelbuild, it will be mistaken
        # for an intermediate modelbuild step and include --close and --initial-transform to the registration.
        # To avoid this, the template file is renamed to another generic filename. 
        command = f'cp {self.inputs.template_anat} {template_folder}/modelbuild_starting_target.nii.gz'
        rc,c_out = run_command(command)
        log.debug(f"The --starting-target template original file is {self.inputs.template_anat}, and was renamed to {template_folder}/modelbuild_starting_target.nii.gz.")

        command = f'QBATCH_SYSTEM={cluster_type} QBATCH_CORES={num_threads} modelbuild.sh \
            --float --average-type median --gradient-step 0.25 --iterations 2 --starting-target {template_folder}/modelbuild_starting_target.nii.gz --stages {stages} \
            --output-dir {template_folder} --sharpen-type unsharp --block --debug {masks} {csv_path}'
        if winsorize_lower_bound>0.0: # only specify an input if they differ from the default
            command+=f" --winsorize_lower_bound {winsorize_lower_bound}"
        if winsorize_upper_bound<1.0: # only specify an input if they differ from the default
            command+=f" --winsorize_upper_bound {winsorize_upper_bound}"
        rc,c_out = run_command(command)

        last_stage = stages.split(',')[-1]

        unbiased_template = template_folder + \
            f'/{last_stage}/1/average/template_sharpen_shapeupdate.nii.gz'
        # verify that all outputs are present
        if not os.path.isfile(unbiased_template):
            raise ValueError(unbiased_template+" doesn't exists.")

        if self.inputs.masking and not (merged_masks[0]=='NULL'):
            unbiased_mask = template_folder + \
                f'/{last_stage}/1/average/mask_shapeupdate.nii.gz'
            # verify that all outputs are present
            if not os.path.isfile(unbiased_mask):
                raise ValueError(unbiased_mask+" doesn't exists.")

        affine_list = []
        warp_list = []
        inverse_warp_list = []
        warped_image_list = []

        import SimpleITK as sitk
        dimension = 3
        identity = sitk.Transform(dimension, sitk.sitkIdentity)

        nlin_out = last_stage=='nlin'
        i = 0
        for file in merged:
            file = str(file)
            filename_template = pathlib.Path(file).name.rsplit(".nii")[0]

            if not nlin_out:
                # create a surrogate transform that takes the place of non-linear since nlin stage is not run
                surrogate_transform_file = f'{template_folder}/{filename_template}_surrogate_identity_transform.mat'
                sitk.WriteTransform(identity, surrogate_transform_file)
                anat_to_unbiased_inverse_warp = surrogate_transform_file
                anat_to_unbiased_warp = surrogate_transform_file
            else:
                anat_to_unbiased_inverse_warp = f'{template_folder}/{last_stage}/1/transforms/{filename_template}_1InverseWarp.nii.gz'
                anat_to_unbiased_warp = f'{template_folder}/{last_stage}/1/transforms/{filename_template}_1Warp.nii.gz'

            if not os.path.isfile(anat_to_unbiased_inverse_warp):
                raise ValueError(
                    anat_to_unbiased_inverse_warp+" file doesn't exists.")
            if not os.path.isfile(anat_to_unbiased_warp):
                raise ValueError(anat_to_unbiased_warp+" file doesn't exists.")
            
            anat_to_unbiased_affine = f'{template_folder}/{last_stage}/1/transforms/{filename_template}_0GenericAffine.mat'
            if not os.path.isfile(anat_to_unbiased_affine):
                raise ValueError(anat_to_unbiased_affine
                                 + " file doesn't exists.")
            warped_image = f'{template_folder}/{last_stage}/1/resample/{filename_template}.nii.gz'
            if not os.path.isfile(warped_image):
                raise ValueError(warped_image
                                 + " file doesn't exists.")
            inverse_warp_list.append(anat_to_unbiased_inverse_warp)
            warp_list.append(anat_to_unbiased_warp)
            affine_list.append(anat_to_unbiased_affine)
            warped_image_list.append(warped_image)
            i += 1

        setattr(self, 'unbiased_template', unbiased_template)
        setattr(self, 'unbiased_mask', unbiased_mask)
        setattr(self, 'affine_list', affine_list)
        setattr(self, 'warp_list', warp_list)
        setattr(self, 'inverse_warp_list', inverse_warp_list)
        setattr(self, 'warped_image_list', warped_image_list)

        return runtime

    def _list_outputs(self):
        return {'unbiased_template': getattr(self, 'unbiased_template'),
                'unbiased_mask': getattr(self, 'unbiased_mask'),
                'affine_list': getattr(self, 'affine_list'),
                'warp_list': getattr(self, 'warp_list'),
                'inverse_warp_list': getattr(self, 'inverse_warp_list'),
                'warped_image_list': getattr(self, 'warped_image_list'), }


def inherit_unbiased_files(RABIES_output_path, opts):
    import os
    from nipype import logging
    log = logging.getLogger('nipype.workflow')
    from rabies.utils import get_workflow_dict
    import pickle
    cli_file = f'{RABIES_output_path}/rabies_preprocess.pkl'
    with open(cli_file, 'rb') as handle:
        inherit_preprocess_opts = pickle.load(handle)

    if inherit_preprocess_opts.commonspace_reg['fast_commonspace']:
        raise ValueError("fast_commonspace was conducted in the previous run; no unbiased template can be inherited.")

    log.warning("--inherit_unbiased_template will be applied. The following options are inherited from the previous run: \n \
            --anatomical_resampling \n \
            --commonspace_reg \n \
            --anat_template \n \
            --brain_mask \n \
            --WM_mask \n \
            --CSF_mask \n \
            --vascular_mask \n")
    
    opts.anatomical_resampling = inherit_preprocess_opts.anatomical_resampling
    opts.commonspace_reg = inherit_preprocess_opts.commonspace_reg
    opts.anat_template = inherit_preprocess_opts.anat_template
    opts.brain_mask = inherit_preprocess_opts.brain_mask
    opts.WM_mask = inherit_preprocess_opts.WM_mask
    opts.CSF_mask = inherit_preprocess_opts.CSF_mask
    opts.vascular_mask = inherit_preprocess_opts.vascular_mask

    preproc_workflow_file = f'{RABIES_output_path}/rabies_preprocess_workflow.pkl'
    node_dict = get_workflow_dict(preproc_workflow_file)

    inherit_dict = {}

    inherit_dict['registration_template'] = node_dict['main_wf.resample_template'].result.outputs.get()['registration_template']
    inherit_dict['registration_mask'] = node_dict['main_wf.resample_template'].result.outputs.get()['registration_mask']
    inherit_dict['commonspace_template'] = node_dict['main_wf.resample_template'].result.outputs.get()['commonspace_template']

    inherit_dict['unbiased_template'] = node_dict['main_wf.commonspace_reg_wf.generate_template'].result.outputs.get()['unbiased_template']
    inherit_dict['unbiased_mask'] = node_dict['main_wf.commonspace_reg_wf.generate_template'].result.outputs.get()['unbiased_mask']

    inherit_dict['unbiased_to_atlas_affine'] = node_dict['main_wf.commonspace_reg_wf.atlas_reg'].result.outputs.get()['affine']
    inherit_dict['unbiased_to_atlas_warp'] = node_dict['main_wf.commonspace_reg_wf.atlas_reg'].result.outputs.get()['warp']
    inherit_dict['unbiased_to_atlas_inverse_warp'] = node_dict['main_wf.commonspace_reg_wf.atlas_reg'].result.outputs.get()['inverse_warp']
    inherit_dict['warped_unbiased'] = node_dict['main_wf.commonspace_reg_wf.atlas_reg'].result.outputs.get()['warped_image']

    # check that the files exist
    for key in list(inherit_dict.keys()):
        try:
            if not os.path.isfile(inherit_dict[key]):
                raise ValueError(f"File not found for {key}. {inherit_dict[key]} was found instead.")
        except:
            if key=='unbiased_mask': 
                if opts.commonspace_reg['masking']: # the file is only present when masking was used
                    raise ValueError(f"File not found for {key}. {inherit_dict[key]} was found instead.")
            elif key=='unbiased_to_atlas_warp' or key=='unbiased_to_atlas_inverse_warp': # if registration was rigid or affine, no warp files
                if not inherit_dict[key]=='NULL':
                    raise ValueError(f"File not found for {key}. {inherit_dict[key]} was found instead.")
            else:
                raise ValueError(f"File not found for {key}. {inherit_dict[key]} was found instead.")
                
    return opts, inherit_dict


import numpy as np
import SimpleITK as sitk
from .analysis_math import vcorrcoef,closed_form


'''
seed-based FC
'''

def seed_based_FC(CR_dict_file, maps_dict_file, seed_name):
    import os
    import numpy as np
    import SimpleITK as sitk
    import pathlib
    from rabies.utils import recover_3D
    from rabies.analysis_pkg.analysis_math import vcorrcoef

    import pickle
    with open(CR_dict_file, 'rb') as handle:
        CR_data_dict = pickle.load(handle)
    with open(maps_dict_file, 'rb') as handle:
        maps_data_dict = pickle.load(handle)
    bold_file = CR_data_dict['bold_file']
    mask_file = maps_data_dict['mask_file']
    timeseries = CR_data_dict['timeseries']
    seed_arr_dict = maps_data_dict['seed_arr_dict']
    roi_mask = seed_arr_dict[seed_name].astype(bool)

    # extract the voxel timeseries within the mask, and take the mean ROI timeseries
    seed_timeseries = timeseries[:,roi_mask].mean(axis=1)

    corrs = vcorrcoef(timeseries.T, seed_timeseries)
    corrs[np.isnan(corrs)] = 0

    corr_map_img = recover_3D(mask_file, corrs)
    filename_split = pathlib.Path(bold_file).name.rsplit(".nii")[0]
    corr_map_file = os.path.abspath(
        filename_split+'_'+seed_name+'_corr_map.nii.gz')
    
    sitk.WriteImage(corr_map_img, corr_map_file)

    # also save the seed timecourse
    import pandas as pd
    seed_timecourse_csv = os.path.abspath(
        filename_split+'_'+seed_name+'_timecourse.csv')
    pd.DataFrame(seed_timeseries).to_csv(seed_timecourse_csv, header=False, index=False)

    return corr_map_file,seed_timecourse_csv


'''
FC matrix
'''


def run_FC_matrix(CR_dict_file,maps_dict_file, figure_format, roi_type='parcellated'):
    import os
    import pandas as pd
    import SimpleITK as sitk
    import numpy as np
    import pathlib  # Better path manipulation
    from rabies.analysis_pkg.analysis_functions import parcellated_FC_matrix, plot_matrix

    import pickle
    with open(CR_dict_file, 'rb') as handle:
        CR_data_dict = pickle.load(handle)
    with open(maps_dict_file, 'rb') as handle:
        maps_data_dict = pickle.load(handle)


    bold_file = CR_data_dict['bold_file']
    filename_split = pathlib.Path(bold_file).name.rsplit(".nii")
    figname = os.path.abspath(filename_split[0]+f'_FC_matrix.{figure_format}')
    
    timeseries = CR_data_dict['timeseries']
    atlas_idx = maps_data_dict['atlas_idx']
    roi_list = maps_data_dict['roi_list']

    if roi_type == 'parcellated':
        corr_matrix,roi_labels = parcellated_FC_matrix(timeseries, atlas_idx, roi_list)
        matrix_df = pd.DataFrame(corr_matrix, index=roi_labels, columns=roi_labels)
    elif roi_type == 'voxelwise':
        corr_matrix = np.corrcoef(timeseries.T)
        matrix_df = pd.DataFrame(corr_matrix)
    else:
        raise ValueError(
            f"Invalid --ROI_type provided: {roi_type}. Must be either 'parcellated' or 'voxelwise.'")
    plot_matrix(figname, corr_matrix)

    data_file = os.path.abspath(filename_split[0]+'_FC_matrix.csv')
    matrix_df.to_csv(data_file, sep=',')
    return data_file, figname


def parcellated_FC_matrix(sub_timeseries, atlas_idx, roi_list):
    
    timeseries_dict = {}
    for i in roi_list:
        roi_mask = np.asarray(atlas_idx == i, dtype=bool)
        # extract the voxel timeseries within the mask, and take the mean ROI timeseries
        timeseries_dict[str(i)] = sub_timeseries[:,roi_mask].mean(axis=1)

    roi_labels = list(timeseries_dict.keys())
    sub_timeseries = []
    for roi in roi_labels:
        sub_timeseries.append(timeseries_dict[roi])
    corr_matrix = np.corrcoef(sub_timeseries)
    return corr_matrix,roi_labels


def plot_matrix(filename, corr_matrix):
    import matplotlib.pyplot as plt
    fig, ax = plt.subplots(1, 1, figsize=(4, 4))
    g = ax.imshow(corr_matrix, cmap='coolwarm', vmax=1, vmin=-1)
    ax.axis('off')
    cbar = plt.colorbar(g, ax=ax, shrink=0.5)
    cbar.set_label('R score', rotation=270, fontsize=10)
    plt.tight_layout()
    plt.savefig(filename, bbox_inches='tight', dpi=150)


'''
ICA
'''


def run_group_ICA(bold_file_list, mask_file, dim, random_seed, background_image, disableMigp=False):
    import os
    import pandas as pd

    # create a filelist.txt
    file_path = os.path.abspath('filelist.txt')
    from rabies.utils import flatten_list
    merged = flatten_list(list(bold_file_list))
    df = pd.DataFrame(data=merged)
    df.to_csv(file_path, header=False, sep=',', index=False)

    from rabies.utils import run_command
    out_dir = os.path.abspath('group_melodic.ica')
    command = f'melodic -i {file_path} -m {mask_file} -o {out_dir} -d {dim} --report --seed={str(random_seed)} --bgimage={background_image}'
    if disableMigp:
        command+=' --disableMigp'
    rc,c_out = run_command(command)
    IC_file = out_dir+'/melodic_IC.nii.gz'
    return out_dir, IC_file


def run_DR_ICA(CR_dict_file, maps_dict_file,network_weighting):
    import os
    import pandas as pd
    import pathlib  # Better path manipulation
    import SimpleITK as sitk
    from rabies.utils import recover_4D
    from rabies.analysis_pkg.analysis_math import dual_regression

    import pickle
    with open(CR_dict_file, 'rb') as handle:
        CR_data_dict = pickle.load(handle)
    with open(maps_dict_file, 'rb') as handle:
        maps_data_dict = pickle.load(handle)
    bold_file = CR_data_dict['bold_file']
    mask_file = maps_data_dict['mask_file']
    timeseries = CR_data_dict['timeseries']
    prior_map_vectors = maps_data_dict['prior_map_vectors']

    filename_split = pathlib.Path(bold_file).name.rsplit(".nii")

    DR = dual_regression(prior_map_vectors, timeseries)

    if network_weighting=='absolute':
        DR_C = DR['C']*DR['S']
    elif network_weighting=='relative':
        DR_C = DR['C']
    else:
        raise 

    dual_regression_timecourse_csv = os.path.abspath(filename_split[0]+'_dual_regression_timecourse.csv')
    pd.DataFrame(DR['W']).to_csv(dual_regression_timecourse_csv, header=False, index=False)

    # save the subjects' IC maps as .nii file
    DR_maps_filename = os.path.abspath(filename_split[0]+'_DR_maps.nii.gz')
    sitk.WriteImage(recover_4D(mask_file, DR_C.T, bold_file), DR_maps_filename)
    return DR_maps_filename, dual_regression_timecourse_csv


from nipype.interfaces.base import (
    traits, TraitedSpec, BaseInterfaceInputSpec,
    File, BaseInterface
)

class NeuralPriorRecoveryInputSpec(BaseInterfaceInputSpec):
    CR_dict_file = File(exists=True, mandatory=True, desc="Dictionary with prepared input matrices from confound correction.")
    maps_dict_file = File(exists=True, mandatory=True, desc="Dictionary with prepared input matrices with brain maps and masks.")
    prior_bold_idx = traits.List(desc="The index for the ICA components that correspond to bold sources.")
    NPR_temporal_comp = traits.Int(
        desc="number of data-driven temporal components to compute.")
    NPR_spatial_comp = traits.Int(
        desc="number of data-driven spatial components to compute.")
    optimize_NPR_dict = traits.Dict(
        desc="Dictionary with options for optimizing NPR convergence.")
    network_weighting = traits.Str(
        desc="Whether to derive absolute or relative (variance-normalized) network maps.")
    figure_format = traits.Str(
        desc="Select file format for figures.")

class NeuralPriorRecoveryOutputSpec(TraitedSpec):
    NPR_prior_timecourse_csv = File(
        exists=True, desc=".csv with timecourses from the fitted prior sources.")
    NPR_extra_timecourse_csv = File(
        exists=True, desc=".csv with timecourses from the scan-specific extra sources.")
    NPR_prior_filename = File(
        exists=True, desc=".nii file with spatial components from the fitted prior sources.")
    NPR_extra_filename = File(
        exists=True, desc=".nii file with spatial components from the scan-specific extra sources.")
    optimize_report = traits.Any(
        exists=False, desc="The NPR optimization report.")


class NeuralPriorRecovery(BaseInterface):
    """

    """

    input_spec = NeuralPriorRecoveryInputSpec
    output_spec = NeuralPriorRecoveryOutputSpec

    def _run_interface(self, runtime):
        import os
        import numpy as np
        import pandas as pd
        import pathlib  # Better path manipulation
        from rabies.utils import recover_4D
        from rabies.analysis_pkg.analysis_math import spatiotemporal_prior_fit

        import pickle
        with open(self.inputs.CR_dict_file, 'rb') as handle:
            CR_data_dict = pickle.load(handle)
        with open(self.inputs.maps_dict_file, 'rb') as handle:
            maps_data_dict = pickle.load(handle)
        bold_file = CR_data_dict['bold_file']
        mask_file = maps_data_dict['mask_file']
        timeseries = CR_data_dict['timeseries']
        prior_map_vectors = maps_data_dict['prior_map_vectors']
        network_weighting=self.inputs.network_weighting

        filename_split = pathlib.Path(
            bold_file).name.rsplit(".nii")

        C_prior = prior_map_vectors[self.inputs.prior_bold_idx,:].T

        optimize_NPR_dict = self.inputs.optimize_NPR_dict
        if optimize_NPR_dict['apply']:
            modeling,_,_,_,optimize_report_fig = spatiotemporal_fit_converge(timeseries,C_prior,
                                        window_size=optimize_NPR_dict['window_size'],
                                        min_prior_corr=optimize_NPR_dict['min_prior_corr'],
                                        diff_thresh=optimize_NPR_dict['diff_thresh'],
                                        max_iter=optimize_NPR_dict['max_iter'], 
                                        compute_max=optimize_NPR_dict['compute_max'], 
                                        gen_report=True)
            optimize_report_file = os.path.abspath(f'{filename_split[0]}_NPR_optimize.{self.inputs.figure_format}')
            optimize_report_fig.savefig(optimize_report_file, bbox_inches='tight')
        else:
            NPR_temporal_comp = self.inputs.NPR_temporal_comp
            NPR_spatial_comp = self.inputs.NPR_spatial_comp
            if NPR_temporal_comp<1: # make sure there is no negative number
                NPR_temporal_comp=0
            if NPR_spatial_comp<1: # make sure there is no negative number
                NPR_spatial_comp=0
            modeling = spatiotemporal_prior_fit(timeseries, C_prior, num_W=NPR_temporal_comp, num_C=NPR_spatial_comp)
            optimize_report_file=None

        # put together the spatial and temporal extra components
        W_extra = np.concatenate((modeling['W_spatial'],modeling['W_temporal']), axis=1)

        if network_weighting=='absolute':
            C_extra = np.concatenate((modeling['C_spatial']*modeling['S_spatial'],
                modeling['C_temporal']*modeling['S_temporal']), axis=1)
            C_fit = modeling['C_fitted_prior']*modeling['S_fitted_prior']
        elif network_weighting=='relative':
            C_extra = np.concatenate((modeling['C_spatial'],
                modeling['C_temporal']), axis=1)
            C_fit = modeling['C_fitted_prior']
        else:
            raise 

        NPR_prior_timecourse_csv = os.path.abspath(filename_split[0]+'_NPR_prior_timecourse.csv')
        pd.DataFrame(modeling['W_fitted_prior']).to_csv(NPR_prior_timecourse_csv, header=False, index=False)

        NPR_extra_timecourse_csv = os.path.abspath(filename_split[0]+'_NPR_extra_timecourse.csv')
        pd.DataFrame(W_extra).to_csv(NPR_extra_timecourse_csv, header=False, index=False)

        NPR_prior_filename = os.path.abspath(filename_split[0]+'_NPR_prior.nii.gz')
        sitk.WriteImage(recover_4D(mask_file,C_fit.T, bold_file), NPR_prior_filename)

        if (self.inputs.NPR_temporal_comp+self.inputs.NPR_spatial_comp)>0:
            NPR_extra_filename = os.path.abspath(filename_split[0]+'_NPR_extra.nii.gz')
            sitk.WriteImage(recover_4D(mask_file,C_extra.T, bold_file), NPR_extra_filename)
        else:
            empty_img = sitk.GetImageFromArray(np.empty([1,1]))
            empty_file = os.path.abspath('empty.nii.gz')
            sitk.WriteImage(empty_img, empty_file)
            NPR_extra_filename = empty_file

        setattr(self, 'NPR_prior_timecourse_csv', NPR_prior_timecourse_csv)
        setattr(self, 'NPR_extra_timecourse_csv', NPR_extra_timecourse_csv)
        setattr(self, 'NPR_prior_filename', NPR_prior_filename)
        setattr(self, 'NPR_extra_filename', NPR_extra_filename)
        setattr(self, 'optimize_report', optimize_report_file)

        return runtime

    def _list_outputs(self):
        return {'NPR_prior_timecourse_csv': getattr(self, 'NPR_prior_timecourse_csv'),
                'NPR_extra_timecourse_csv': getattr(self, 'NPR_extra_timecourse_csv'),
                'NPR_prior_filename': getattr(self, 'NPR_prior_filename'),
                'NPR_extra_filename': getattr(self, 'NPR_extra_filename'),
                'optimize_report': getattr(self, 'optimize_report'),
                }


def eval_convergence(prior_corr_list,fit_diff_list,window_size=5,min_prior_corr=0.5,diff_thresh=0.04):
    if len(fit_diff_list)<window_size:
        return None
    for i in range(1,len(fit_diff_list)-window_size+1): # we start at 1 because idx 0 is nan
        if (np.array(prior_corr_list[i-1])<min_prior_corr).max():
            continue # if the preceding iteration does not pass corr threshold, search for another window
        window = np.array(fit_diff_list[i:i+window_size])
        if not (window>diff_thresh).max(): # if all diff are below threshold, return the first index
            return i-1 # we take -1 because the iteration preceding the window is ideal
    return None


def spatiotemporal_fit_converge(X,C_prior,window_size=5,min_prior_corr=0.5,diff_thresh=0.03,max_iter=20, compute_max=False, gen_report=False):
    '''
    NPR is conducted while incrementally increasing the number of components (alternating 
    between a spatial and a temporal component). 
    
    Convergence criterion 1: Iterations continue until the correlation between the fitted
    component and the prior does not reach the specified minimum.
    
    Convergence Criterion 2: At each iteration, the difference between the previous and new 
    output is evaluated (0=perfectly correlated; 1=uncorrelated). The forming set of 
    successive iterations (within a certain window length) is evaluated, and when a set 
    respects the convergence threshold for each iteration within the window, the iteration 
    preceding that window is selected as optimal output. We take the iteration preceding 
    the  window, as this corresponds to the last iteration which generated changes above 
    threshold. The sliding-window approach is employed to prevent falling within a local 
    minima, when further ameliorations may be possible with further iterations.
    
    When multiple priors are fitted, they are all simultaneously subjected to the evaluation
    of convergence, and as long as one prior fit does not meet the thresholds, iterations
    continue.
    
    X: timeseries of shape time by voxel
    C_prior: set of priors of shape voxel by num priors
    window_size: the size of the sliding window
    min_prior_corr: the minimum correlation with the prior
    diff_thresh: the convergence threshold
    max_iter: maximum number of components to derive
    '''
    
    from rabies.analysis_pkg.analysis_math import spatiotemporal_prior_fit

    num_prior = C_prior.shape[1]
    num_list=range(0,max_iter)

    fit_list=[]
    # first entry is NaN, since there is no comparison to make
    fit_diff_list=[np.repeat(np.nan,num_prior)] 
    NPR_dict_list=[]
    prior_corr_list=[]
    for num in num_list:
        num_C = int(num/2)
        if num%2:
            num_W = num_C+1
        else:
            num_W = num_C

        NPR_dict = spatiotemporal_prior_fit(X, C_prior, num_W=num_W, num_C=num_C)
        NPR_dict_list.append(NPR_dict)
        prior_corr_list.append(NPR_dict['corr_list'])
        C_fit = NPR_dict['C_fitted_prior']
        C_fit /= np.sqrt((C_fit ** 2).sum(axis=0))
        
        if len(fit_list)>0:
            ##### evaluate convergence
            C = C_fit
            C_prev = fit_list[-1]
            # each prior has a separate lim evaluation
            lim = np.abs(np.abs((C * C_prev).sum(axis=0)) - 1)
            fit_diff_list.append(lim)

        convergence_idx = eval_convergence(prior_corr_list,fit_diff_list,window_size=window_size,min_prior_corr=min_prior_corr,diff_thresh=diff_thresh)
        if not compute_max and not convergence_idx is None:
            if gen_report:
                fig = generate_convergence_report(convergence_idx,fit_diff_list,prior_corr_list,min_prior_corr=min_prior_corr,diff_thresh=diff_thresh)
            return NPR_dict_list[convergence_idx],convergence_idx,fit_diff_list,prior_corr_list,fig
        
        fit_list.append(C_fit)

    if gen_report:
        fig = generate_convergence_report(convergence_idx,fit_diff_list,prior_corr_list,min_prior_corr=min_prior_corr,diff_thresh=diff_thresh)
    if convergence_idx is None: # if there was no convergence, the last iteration is outputed
        convergence_idx = -1
    return NPR_dict_list[convergence_idx],None,fit_diff_list,prior_corr_list,fig


def generate_convergence_report(convergence_idx,fit_diff_list,prior_corr_list,min_prior_corr=0.5,diff_thresh=0.03):
    import matplotlib.pyplot as plt

    num_list=range(len(fit_diff_list))
    num_priors = len(prior_corr_list[0])
    
    fig,axes=plt.subplots(1,2,figsize=(10,5))
    
    axes[0].plot(num_list,prior_corr_list)
    axes[1].plot(num_list,fit_diff_list)
    if not convergence_idx is None:
        axes[0].scatter([num_list[convergence_idx]]*num_priors,prior_corr_list[convergence_idx], color='r', marker='*', s=80)
        axes[1].scatter([num_list[convergence_idx]]*num_priors,fit_diff_list[convergence_idx], color='r', marker='*', s=80)
        
    axes[0].set_ylim([0,1])
    axes[0].set_ylabel('Correlation with prior', color='white', fontsize=15)
    axes[1].set_ylim([0,min(diff_thresh*4,1)])
    axes[1].set_ylabel('Difference from previous', color='white', fontsize=15)
    
    axes[0].set_xlabel('Number of non-prior components', color='white', fontsize=15)
    axes[0].set_xlim([-1,max(num_list)+1])
    axes[1].set_xlabel('Number of non-prior components', color='white', fontsize=15)
    axes[1].set_xlim([-1,max(num_list)+1])
    
    axes[0].plot([-1,max(num_list)+1],[min_prior_corr,min_prior_corr], color='lightgray', linestyle='--', linewidth=2)
    axes[1].plot([-1,max(num_list)+1],[diff_thresh,diff_thresh], color='lightgray', linestyle='--', linewidth=2)
    
    legend_labels = [f'Prior {i+1}' for i in range(num_priors)]
    axes[0].legend(legend_labels, fontsize=12, loc='lower right')
    axes[1].legend(legend_labels, fontsize=12, loc='upper right')
    plt.tight_layout()
    return fig


import os
import pathlib  # Better path manipulation
import SimpleITK as sitk
import numpy as np
from nipype.interfaces.base import (
    traits, TraitedSpec, BaseInterfaceInputSpec,
    File, BaseInterface
)

######################
#IMAGE MANIPULATION
######################


def get_sitk_header(file_path): # read a nifti header without loading the data
    reader = sitk.ImageFileReader()
    reader.SetFileName(file_path)
    reader.ReadImageInformation()
    return reader


def get_geometry_header(img):
    # this creates an empty SITK image that carries information about geometry, but not the voxel array size
    # IMPORTANTLY: this means that header.GetSize() or CopyInformation(header) is inaccurate.
    header_geo = sitk.Image([1]*img.GetDimension(), img.GetPixelID())
    header_geo.SetSpacing(img.GetSpacing())
    header_geo.SetOrigin(img.GetOrigin())
    header_geo.SetDirection(img.GetDirection())
    for k in img.GetMetaDataKeys(): # metadata is also copied in case
        header_geo.SetMetaData(k, img.GetMetaData(k))
    return header_geo


def recover_3D(mask_img, vector_map):
    if isinstance(mask_img, sitk.Image):
        pass
    elif os.path.isfile(mask_img):
        mask_img = sitk.ReadImage(mask_img)
    else:
        raise ValueError(f"{mask_img} is neither an SITK image nor a file path.")
    brain_mask = sitk.GetArrayFromImage(mask_img)
    volume_indices=brain_mask.astype(bool)
    volume=np.zeros(brain_mask.shape)
    volume[volume_indices]=vector_map
    volume_img = copyInfo_3DImage(sitk.GetImageFromArray(
        volume, isVector=False), mask_img)
    return volume_img


def recover_4D(mask_img, vector_maps, ref_4d):
    if isinstance(mask_img, sitk.Image):
        pass
    elif os.path.isfile(mask_img):
        mask_img = sitk.ReadImage(mask_img)
    else:
        raise ValueError(f"{mask_img} is neither an SITK image nor a file path.")
 
    if isinstance(ref_4d, sitk.Image):
        pass
    elif isinstance(ref_4d, sitk.ImageFileReader):
        pass
    elif os.path.isfile(ref_4d):
        ref_4d = get_sitk_header(ref_4d)
    else:
        raise ValueError(f"{ref_4d} is neither an SITK image nor a file path.")
    
    #vector maps of shape num_volumeXnum_voxel
    brain_mask = sitk.GetArrayFromImage(mask_img)
    volume_indices=brain_mask.astype(bool)
    shape=(vector_maps.shape[0],brain_mask.shape[0],brain_mask.shape[1],brain_mask.shape[2])
    volumes=np.zeros(shape)
    for i in range(vector_maps.shape[0]):
        volume=volumes[i,:,:,:]
        volume[volume_indices]=vector_maps[i,:]
        volumes[i,:,:,:]=volume
    volume_img = copyInfo_4DImage(sitk.GetImageFromArray(
        volumes, isVector=False), mask_img, ref_4d)
    return volume_img


def resample_image_spacing(image, output_spacing):
    dimension = 3
    identity = sitk.Transform(dimension, sitk.sitkIdentity)

    # Compute grid size based on the physical size and spacing.
    input_size = image.GetSize()
    sampling_ratio = np.asarray(image.GetSpacing())/np.asarray(output_spacing)
    output_size = [int(input_size[0]*sampling_ratio[0]), int(input_size[1]
                                                             * sampling_ratio[1]), int(input_size[2]*sampling_ratio[2])]

    # set default threader to platform to avoid freezing with MultiProc https://github.com/SimpleITK/SimpleITK/issues/1239
    sitk.ProcessObject_SetGlobalDefaultThreader('Platform')
    # the resampling is done with BSpline, the interpolator order is 3
    # other BSpline options are blurry, see https://discourse.itk.org/t/resample-produces-blurry-results-when-just-cropping/4473
    resampled_image = sitk.Resample(image, output_size, identity, sitk.sitkBSpline,
                                    image.GetOrigin(), output_spacing, image.GetDirection())
    # clip potential negative values
    array = sitk.GetArrayFromImage(resampled_image)
    array[(array < 0).astype(bool)] = 0
    pos_resampled_image = sitk.GetImageFromArray(array, isVector=False)
    pos_resampled_image.CopyInformation(resampled_image)
    return pos_resampled_image


def resample_image_spacing_4d(image_4d, output_spacing, clip_negative=True): 
    # output_spacing should be specifying the resolution of the 3 spatial dimensions in mm
    if not image_4d.GetDimension()==4:
        raise ValueError("The image must be 4 dimensional.")
    
    # Compute grid size based on the physical size and spacing.
    input_size = image_4d.GetSize()
    sampling_ratio = np.asarray(image_4d.GetSpacing())[:3]/np.asarray(output_spacing)
    output_size = [int(input_size[0]*sampling_ratio[0]), int(input_size[1]
                                                                * sampling_ratio[1]), int(input_size[2]*sampling_ratio[2])]

    origin = image_4d.GetOrigin()[:3]

    direction_4d = image_4d.GetDirection()
    direction_3d = tuple(direction_4d[:3]+direction_4d[4:7]+direction_4d[8:11])

    dimension = 3
    identity = sitk.Transform(dimension, sitk.sitkIdentity)
    # set default threader to platform to avoid freezing with MultiProc https://github.com/SimpleITK/SimpleITK/issues/1239
    sitk.ProcessObject_SetGlobalDefaultThreader('Platform')
    resampled_list=[]
    for i in range(input_size[3]):
        resampled_image = sitk.Resample(image_4d[:,:,:,i], output_size, identity, sitk.sitkLinear,
                                        origin, output_spacing, direction_3d)
        resampled_list.append(resampled_image)
    combined = sitk.JoinSeries(resampled_list) 
    resampled_4d = copyInfo_4DImage(combined, resampled_image, image_4d)

    if clip_negative:
        # clip potential negative values
        array = sitk.GetArrayFromImage(resampled_4d)
        array[(array < 0).astype(bool)] = 0
        pos_resampled_image = sitk.GetImageFromArray(array, isVector=False)
        pos_resampled_image.CopyInformation(resampled_4d)
        resampled_4d = pos_resampled_image
    return resampled_4d


def copyInfo_4DImage(image_4d, ref_3d, ref_4d):
    # function to establish metadata of an input 4d image. The ref_3d will provide
    # the information for the first 3 dimensions, and the ref_4d for the 4th.
    if ref_3d.GetDimension() == 4:
        image_4d.SetSpacing(
            tuple(list(ref_3d.GetSpacing()[:3])+[ref_4d.GetSpacing()[3]]))
        image_4d.SetOrigin(
            tuple(list(ref_3d.GetOrigin()[:3])+[ref_4d.GetOrigin()[3]]))
        dim_3d = list(ref_3d.GetDirection())
        dim_4d = list(ref_4d.GetDirection())
        image_4d.SetDirection(
            tuple(dim_3d[:3]+[dim_4d[3]]+dim_3d[4:7]+[dim_4d[7]]+dim_3d[8:11]+dim_4d[11:]))
    elif ref_3d.GetDimension() == 3:
        image_4d.SetSpacing(
            tuple(list(ref_3d.GetSpacing())+[ref_4d.GetSpacing()[3]]))
        image_4d.SetOrigin(
            tuple(list(ref_3d.GetOrigin())+[ref_4d.GetOrigin()[3]]))
        dim_3d = list(ref_3d.GetDirection())
        dim_4d = list(ref_4d.GetDirection())
        image_4d.SetDirection(
            tuple(dim_3d[:3]+[dim_4d[3]]+dim_3d[3:6]+[dim_4d[7]]+dim_3d[6:9]+dim_4d[11:]))
    else:
        raise ValueError('Unknown reference image dimensions.')
    return image_4d


def copyInfo_3DImage(image_3d, ref_3d):
    if ref_3d.GetDimension() == 4:
        image_3d.SetSpacing(ref_3d.GetSpacing()[:3])
        image_3d.SetOrigin(ref_3d.GetOrigin()[:3])
        dim_3d = list(ref_3d.GetDirection())
        image_3d.SetDirection(tuple(dim_3d[:3]+dim_3d[4:7]+dim_3d[8:11]))
    elif ref_3d.GetDimension() == 3:
        image_3d.SetSpacing(ref_3d.GetSpacing())
        image_3d.SetOrigin(ref_3d.GetOrigin())
        image_3d.SetDirection(ref_3d.GetDirection())
    else:
        raise ValueError('Unknown reference image dimensions.')
    return image_3d


class ResampleVolumesInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="Input 3D or 4D file to resample")
    ref_file = File(exists=True, mandatory=True,
                    desc="The reference 3D space to which the EPI will be warped.")
    transforms_3d_files = traits.List(desc="List of transform files to apply to every volume.")
    inverses_3d = traits.List(
        desc="Define whether some transforms must be inverse, with a boolean list where true defines inverse e.g.[0,1,0]")
    apply_motcorr = traits.Bool(
        default=True, desc="Whether to apply motion realignment, only for 4D file.")
    motcorr_params = File(
        exists=True, desc="xforms from head motion estimation .csv file")
    resampling_dim = traits.Str(
        desc="Specify the image dimension of post-resampling.")
    interpolation = traits.Int(
        desc="An SITK interpolator for resampling.")
    name_source = File(exists=True, mandatory=True,
                         desc='a Nifti file from which the header should be copied')
    clip_negative = traits.Bool(
        desc="Whether to clip out negative values after resampling.")
    rabies_data_type = traits.Int(mandatory=True,
                                  desc="Integer specifying SimpleITK data type.")


class ResampleVolumesOutputSpec(TraitedSpec):
    resampled_file = File(exists=True, desc='output 4D resampled file')


class ResampleVolumes(BaseInterface):
    """
    This interface will apply a set of transforms to an input 4D EPI as well as motion realignment if specified.
    Susceptibility distortion correction can be applied through the provided transforms. A list of the corrected
    single volumes will be provided as outputs, and these volumes require to be merged to recover timeseries.
    """

    input_spec = ResampleVolumesInputSpec
    output_spec = ResampleVolumesOutputSpec

    def _run_interface(self, runtime):
        # set default threader to platform to avoid freezing with MultiProc https://github.com/SimpleITK/SimpleITK/issues/1239
        sitk.ProcessObject_SetGlobalDefaultThreader('Platform')        
        # avoid crashes if receiving an empty file
        if pathlib.Path(self.inputs.in_file).name=='empty.nii.gz':
            setattr(self, 'resampled_file', self.inputs.in_file)
            return runtime

        # prepare output name
        filename_split = pathlib.Path(
            self.inputs.name_source).name.rsplit(".nii")
        resampled_file = os.path.abspath(
            f"{filename_split[0]}_resampled.nii.gz")

        ref_img = sitk.ReadImage(self.inputs.ref_file, self.inputs.rabies_data_type)

        # preparing the resampling dimensions of the reference space
        if not self.inputs.resampling_dim == 'ref_file': # with 'ref_file' the reference file is unaltered, and thus directly defines the commonspace resolution
            if self.inputs.resampling_dim == 'inputs_defined':
                spacing = get_sitk_header(self.inputs.in_file).GetSpacing()[:3]
            else:
                shape = self.inputs.resampling_dim.split('x')
                spacing = (float(shape[0]), float(shape[1]), float(shape[2]))
            ref_img = resample_image_spacing(ref_img, spacing)

        # replace undefined inputs for an empty list to avoid crashes
        from nipype.interfaces.base import isdefined
        transforms_3d_files = self.inputs.transforms_3d_files if isdefined(self.inputs.transforms_3d_files) else []
        inverses_3d = self.inputs.inverses_3d if isdefined(self.inputs.inverses_3d) else []
        n_procs = int(os.environ['RABIES_ITK_NUM_THREADS']) if "RABIES_ITK_NUM_THREADS" in os.environ else os.cpu_count() # default to number of CPUs

        motcorr_params_file = self.inputs.motcorr_params if self.inputs.apply_motcorr else None

        resampled_img = resample_volumes(self.inputs.in_file, ref_img, transforms_3d_files, inverses_3d, 
                                         motcorr_params_file = motcorr_params_file, interpolation=self.inputs.interpolation, 
                                         rabies_data_type=self.inputs.rabies_data_type, clip_negative=self.inputs.clip_negative, 
                                         n_procs=n_procs)
        sitk.WriteImage(resampled_img, resampled_file)

        setattr(self, 'resampled_file', resampled_file)
        return runtime

    def _list_outputs(self):
        return {'resampled_file': getattr(self, 'resampled_file')}


def resample_volumes(in_img, in_ref, transforms_3d_files = None, inverses_3d = None, motcorr_params_file = None, interpolation=sitk.sitkLinear, rabies_data_type=8, clip_negative=False, n_procs=None):
    import SimpleITK as sitk
    from simpleitk_timeseries_motion_correction.apply_transforms import read_transforms_from_csv, resample_volume

    if transforms_3d_files is None:
        transforms_3d_files = []
    if inverses_3d is None:
        inverses_3d = []
    if n_procs is None:
        n_procs = os.cpu_count()

    # the input can be either a nifti file or an SITK image
    if isinstance(in_img, sitk.Image):
        orig_img = in_img
    elif os.path.isfile(in_img):
        orig_img = sitk.ReadImage(in_img, rabies_data_type)
    if isinstance(in_ref, sitk.Image):
        ref_img = in_ref
    elif os.path.isfile(in_ref):
        ref_img = sitk.ReadImage(in_ref, rabies_data_type)
    del in_img, in_ref # release from memory

    transforms_3d_l = load_sitk_transforms(transforms_3d_files, inverses_3d)

    if orig_img.GetDimension()==3:
        composite = sitk.CompositeTransform(3)
        for transform in transforms_3d_l:
            composite.AddTransform(transform)
        resampled_img = resample_volume(orig_img, ref_img, composite, interpolation=interpolation, clip_negative=clip_negative, extrapolator=False)

    elif orig_img.GetDimension()==4:
        import concurrent.futures
        from tqdm import tqdm
        if motcorr_params_file is not None:
            hmc_transforms_l = read_transforms_from_csv(csv_file=motcorr_params_file)

        # create a resampling function that can create the composite transform on a per iteration basis to avoid memory overloads
        def resample_volume_composite(volume, reference, composite_transform_list, interpolation, clip_negative=False, extrapolator=False):
            composite = sitk.CompositeTransform(3)
            for transform in composite_transform_list:
                composite.AddTransform(transform)        
            return resample_volume(volume, reference, composite, interpolation, clip_negative, extrapolator)

        num_volumes = orig_img.GetSize()[3]
        # Convert to a list of 3D volumes from input to process in parallel
        input_volumes = []
        size_4d = orig_img.GetSize()
        for i in range(num_volumes):
            extractor = sitk.ExtractImageFilter()
            extractor.SetSize([size_4d[0], size_4d[1], size_4d[2], 0])
            extractor.SetIndex([0, 0, 0, i])
            input_volumes.append(extractor.Execute(orig_img))

        # before releasing orig_img, copy the header into an empty image
        header_geo = get_geometry_header(orig_img)
        del orig_img, extractor # free memory

        resampled_volumes = [None] * num_volumes

        # Parallel processing
        with concurrent.futures.ThreadPoolExecutor(max_workers=n_procs) as executor:    
            with tqdm(total=num_volumes) as pbar:
                futures = {}
                for i in range(num_volumes):
                    if motcorr_params_file is not None:
                        composite_transform_list = [hmc_transforms_l[i]] + transforms_3d_l
                    else:
                        composite_transform_list = transforms_3d_l
                        
                    future = executor.submit(
                        resample_volume_composite,
                        volume=input_volumes[i],
                        reference=ref_img,
                        composite_transform_list=composite_transform_list,
                        interpolation=interpolation,
                        clip_negative=clip_negative,
                        extrapolator=False,
                    )
                    futures[future] = i
                
                for future in concurrent.futures.as_completed(futures):
                    i = futures[future]
                    try:
                        resampled_volumes[i] = future.result()
                    except Exception as e:
                        print(f"Error processing volume {i}: {e}")
                        # Might want to abort or insert blank?
                        # For now re-raise to fail fast
                        raise e
                    pbar.update(1)
        resampled_img = sitk.JoinSeries(resampled_volumes)
        resampled_img = copyInfo_4DImage(resampled_img, resampled_img, header_geo) # only copy original header for 4th dimension
    else:
        raise ValueError(f"Input image must be 3D or 4D. Got {orig_img.GetDimension()}D.")
    
    return resampled_img


def load_sitk_transforms(transforms_files, inverses):
    # input transforms are listed from last to first applied as per antsApplyTransforms convention; this need to be inverted since SITK applies from first to last
    transforms_files = transforms_files[::-1]
    inverses = inverses[::-1]

    # load as a list of SITK transforms
    transforms_l = []
    for transform_file,inverse in zip(transforms_files,inverses):
        if transform_file[-4:]=='.mat':
            transform = sitk.ReadTransform(transform_file)
            if inverse==1:
                transform = transform.GetInverse()
        elif transform_file[-7:]=='.nii.gz':
            # Read the displacement field image and convert to a transform
            transform = sitk.DisplacementFieldTransform(sitk.ReadImage(
                transform_file,
                sitk.sitkVectorFloat64
            ))
        elif transform_file=='NULL':
            continue # don't append if NULL, which is a possible RABIES input
        else:
            raise ValueError(f"Transform files must end with .mat or .nii.gz. Cannot read input {transform_file}.")
        transforms_l.append(transform)
    return transforms_l


class ResampleMaskInputSpec(BaseInterfaceInputSpec):
    mask_file = File(exists=True, mandatory=True,
                desc="Input mask.")
    ref_file = File(exists=True, mandatory=True,
                   desc="Target file defining resampling space.")
    transforms = traits.List(desc="List of transforms to resample the mask.")
    inverses = traits.List(
        desc="Define whether some transforms must be inverse, with a boolean list where true defines inverse e.g.[0,1,0]")
    name_suffix = traits.Str(desc="Suffix added at the output file.")
    name_source = File(exists=True, mandatory=True,
                       desc='Reference file for prefix of the output file.')


class ResampleMaskOutputSpec(TraitedSpec):
    resampled_file = traits.File(desc="The resampled mask file.")


class ResampleMask(BaseInterface):

    input_spec = ResampleMaskInputSpec
    output_spec = ResampleMaskOutputSpec

    def _run_interface(self, runtime):
        import os
        import pathlib  # Better path manipulation
        filename_split = pathlib.Path(
            self.inputs.name_source).name.rsplit(".nii")

        if self.inputs.name_suffix is None:
            new_mask_path = os.path.abspath(
                f'{filename_split[0]}_resampled.nii.gz')
        else:
            new_mask_path = os.path.abspath(f'{filename_split[0]}_{self.inputs.name_suffix}.nii.gz')

        # replace undefined inputs for an empty list to avoid crashes
        from nipype.interfaces.base import isdefined
        transforms = self.inputs.transforms if isdefined(self.inputs.transforms) else []
        inverses = self.inputs.inverses if isdefined(self.inputs.inverses) else []

        antsApplyTransforms(transforms = transforms, inverses = inverses, 
                        input_image = self.inputs.mask_file, ref_image = self.inputs.ref_file, output_filename = new_mask_path, interpolation='GenericLabel', rabies_data_type=sitk.sitkInt16, clip_negative=False)

        setattr(self, 'resampled_file', new_mask_path)
        return runtime

    def _list_outputs(self):
        return {'resampled_file': getattr(self, 'resampled_file')}


def antsApplyTransforms(transforms, inverses, input_image, ref_image, output_filename, interpolation, rabies_data_type=8, clip_negative=False):
    # tranforms is a list of transform files, set in order of call within antsApplyTransforms
    transform_string = ""
    for transform, inverse in zip(transforms, inverses):
        if transform=='NULL':
            continue
        elif bool(inverse):
            transform_string += f"-t [{transform},1] "
        else:
            transform_string += f"-t {transform} "

    command = f'antsApplyTransforms -i {input_image} {transform_string}-n {interpolation} -r {ref_image} -o {output_filename} -v'
    rc,c_out = run_command(command)

    if clip_negative or rabies_data_type is not None: # rabies_data_type can be set to None to avoid re-writing the image
        if rabies_data_type is not None:
            # need to reload/save the image to switch to rabies_data_type
            resampled_img = sitk.ReadImage(output_filename, rabies_data_type)
        else:
            resampled_img = sitk.ReadImage(output_filename)
        if clip_negative:
            # clip potential negative values
            array = sitk.GetArrayFromImage(resampled_img)
            array[(array < 0).astype(bool)] = 0
            pos_resampled_image = sitk.GetImageFromArray(array, isVector=False)
            pos_resampled_image.CopyInformation(resampled_img)
            resampled_img = pos_resampled_image
        sitk.WriteImage(resampled_img, output_filename)
    if not os.path.isfile(output_filename):
        raise ValueError(
            "Missing output image. Transform call failed: "+command)


######################
#GENERAL
######################

def run_command(command, verbose = False):
    # Run command and collect stdout
    # http://blog.endpoint.com/2015/01/getting-realtime-output-using-python.html # noqa
    from nipype import logging
    log = logging.getLogger('nipype.workflow')
    log.debug('Running: '+command)

    import subprocess
    try:
        if os.environ.get('RABIES_ITK_THREADS_STATE') == 'ON':
            itk_threads = os.environ.get('RABIES_ITK_NUM_THREADS')
            if itk_threads is None:
                raise ValueError("RABIES_ITK_NUM_THREADS must be set when enabling ITK threading control.")
            os.environ['ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS'] = itk_threads        
        process = subprocess.run(
            command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
            check=True,
            shell=True,
            env=os.environ,
            )
    except Exception as e:
        log.warning(e.output.decode("utf-8"))
        raise

    c_out = process.stdout.decode("utf-8")
    if not c_out == '':
        if verbose:
            log.info(c_out)
        else:
            log.debug(c_out)
    if process.stderr is not None:
        if verbose:
            log.info(process.stderr)
        else:
            log.warning(process.stderr)
    rc = process.returncode
    if len(c_out)>0 and c_out[-1]=='\n': # remove the extra break point, can affect string construction
        c_out = c_out[:-1] 
    return rc,c_out


def flatten_list(l):
    if type(l) == list:
        flattened = []
        for e in l:
            e = flatten_list(e)
            if type(e) == list:
                flattened += e
            else:
                flattened += [e]
        return flattened
    else:
        return l


def filter_scan_exclusion(exclusion_list, split_name):
    # the function removes a list of scan IDs from split_name
    
    # exclusion_list: the input provided by the user
    # split_name: a list of all scan IDs that were found
    
    import numpy as np
    import pandas as pd
    if os.path.isfile(os.path.abspath(exclusion_list[0])):
        updated_split_name=[]
        if not '.nii' in pathlib.Path(exclusion_list[0]).name:
            # read the file as a .txt
            exclusion_list = np.array(pd.read_csv(os.path.abspath(exclusion_list[0]), header=None)).flatten()
        for split in split_name:
            exclude = False
            for scan in exclusion_list:
                if split in scan:
                    exclude = True
            if not exclude:
                updated_split_name.append(split)
    elif exclusion_list[0]=='none':
        updated_split_name = split_name
    else:
        raise ValueError(f"The --exclusion_ids {exclusion_list} input had improper format. It must the full path to a .txt or .nii files.")
    
    if len(updated_split_name)==0:
        raise ValueError(f"""
            No scans are left after scan exclusion!
            """)

    return updated_split_name


def filter_scan_inclusion(inclusion_list, split_name):
    # the function will update the list of scan IDs in split_name to correspond to inclusion/exclusion list
    
    # inclusion_list: the input provided by the user
    # split_name: a list of all scan IDs that were found
    
    import numpy as np
    import pandas as pd
    if os.path.isfile(os.path.abspath(inclusion_list[0])):
        updated_split_name=[]
        if '.nii' in pathlib.Path(inclusion_list[0]).name:
            for scan in inclusion_list:
                updated_split_name.append(find_split(scan, split_name))
        else:
            # read the file as a .txt
            inclusion_list = np.array(pd.read_csv(os.path.abspath(inclusion_list[0]), header=None)).flatten()
            for scan in inclusion_list:
                updated_split_name.append(find_split(scan, split_name))
    elif inclusion_list[0]=='all':
        updated_split_name = split_name
    else:
        raise ValueError(f"The --inclusion_ids {inclusion_list} input had improper format. It must the full path to a .txt or .nii files, or 'all' to keep all scans.")
    return updated_split_name


def find_split(scan, split_name):
    for split in split_name:
        if split in scan:
            return split
    raise ValueError(f"No previous file name is matching {scan}")


######################
#FUNCTIONS TO READ WORKFLOW GRAPH
######################

def fill_split_dict(d, output_bold, split_name, split_dict, keys, node_dict, match_targets):
    if isinstance(d, dict):
        for key in list(d.keys()):
            fill_split_dict(d[key], output_bold, split_name, split_dict, keys+[key], node_dict, match_targets)
    else:
        f = d.result.outputs.get()[output_bold]
        split = pathlib.Path(f).name.rsplit(".nii")[0]
        split_name.append(split)
        split_dict[split]={}
        target_list = list(match_targets.keys())
        for target in target_list:
            if match_targets[target] is None: # certain target may not have been generated, and needs to be attributed None
                split_dict[split][target] = None
            else:
                [unit, output] = match_targets[target]
                node = retrieve_node(node_dict[unit], keys)
                split_dict[split][target] = node.result.outputs.get()[output]
        
def retrieve_node(d, keys):
    if isinstance(d, dict):
        return retrieve_node(d[keys[0]], keys[1:])
    else:
        return d


def get_workflow_dict(workflow_file):
    import pickle
    with open(workflow_file, 'rb') as handle:
        graph = pickle.load(handle)
    
    node_list = list(graph.nodes)
    node_dict = {}
    for node in node_list:
        key_l = [node.fullname]+node.parameterization
        fill_node_dict(node_dict, key_l, node)
    return node_dict


def fill_node_dict(d, key_l, e):
    if len(key_l)>0:
        key = key_l[0]
        if not (key in list(d.keys())):
            d[key] = {}
        d[key] = fill_node_dict(d[key], key_l[1:], e)
        return d
    else:
        return e


######################
#DEBUGGING
######################

def generate_token_data(tmppath, number_scans):
    # this function generates fake scans at low resolution for quick testing and debugging

    os.makedirs(tmppath+'/inputs', exist_ok=True)

    from . import run_main
    template = run_main.DSURQE_ANAT
    mask = run_main.DSURQE_MASK
    labels_file = run_main.DSURQE_LABELS
    melodic_file = run_main.DSURQE_ICA

    spacing = (float(1), float(1), float(1))  # resample to 1mmx1mmx1mm
    resampled_template = resample_image_spacing(sitk.ReadImage(template), spacing)
    # generate template masks
    resampled_mask = resample_image_spacing(sitk.ReadImage(mask), spacing)
    array = sitk.GetArrayFromImage(resampled_mask)
    array[array < 1] = 0
    array[array > 1] = 1
    binarized = sitk.GetImageFromArray(array, isVector=False)
    binarized.CopyInformation(resampled_mask)
    sitk.WriteImage(binarized, tmppath+'/inputs/token_mask.nii.gz')
    array[:, :, :6] = 0
    binarized = sitk.GetImageFromArray(array, isVector=False)
    binarized.CopyInformation(resampled_mask)
    sitk.WriteImage(binarized, tmppath+'/inputs/token_mask_half.nii.gz')
    resampled_labels = resample_image_spacing(sitk.ReadImage(labels_file), spacing)
    sitk.WriteImage(resampled_labels, tmppath+'/inputs/token_labels.nii.gz')

    # generate fake scans from the template
    array = sitk.GetArrayFromImage(resampled_template)
    array_4d = np.repeat(array[np.newaxis, :, :, :], 15, axis=0)
    
    melodic_img = sitk.ReadImage(melodic_file)
    # create a new melodic with just 2 networks for low-dimensional dual regression
    melodic_networks = sitk.JoinSeries([melodic_img[:,:,:,5],melodic_img[:,:,:,19]]) 
    sitk.WriteImage(melodic_networks, tmppath+'/inputs/melodic_networks.nii.gz')
    
    dim = melodic_img.GetSize()[-1]
    for i in range(dim):
        IC_map = sitk.GetArrayFromImage(sitk.Resample(melodic_img[:,:,:,i], resampled_template))
        time = np.random.normal(0, array_4d.mean()/100, array_4d.shape[0]) # network timecourse; scale is 1% of image intensity
        array_4d+=(np.repeat(IC_map[np.newaxis, :, :, :], 15, axis=0).T*time).T

    for i in range(number_scans):
        # generate anatomical scan
        sitk.WriteImage(resampled_template, tmppath+f'/inputs/sub-token{i+1}_T1w.nii.gz')
        # generate functional scan
        array_4d_ = array_4d + np.random.normal(0, array_4d.mean()/ 100, array_4d.shape)  # add gaussian noise; scale is 1% of the mean intensity of the template
        sitk.WriteImage(sitk.GetImageFromArray(array_4d_, isVector=False),
                        tmppath+f'/inputs/sub-token{i+1}_bold.nii.gz')

        # necessary to read matrix orientation properly at the analysis stage
        sitk.WriteImage(copyInfo_4DImage(sitk.ReadImage(tmppath+f'/inputs/sub-token{i+1}_bold.nii.gz'), sitk.ReadImage(tmppath
                        + f'/inputs/sub-token{i+1}_T1w.nii.gz'), sitk.ReadImage(tmppath+f'/inputs/sub-token{i+1}_bold.nii.gz')), tmppath+f'/inputs/sub-token{i+1}_bold.nii.gz')
